import re
from unittest.mock import patch

import numpy as np
import pytest

from netcdf_scm.weights import (
    AreaSurfaceFractionWeightCalculator,
    InvalidWeights,
    get_land_weights,
    get_nh_weights,
    get_weights_for_area,
    multiply_weights,
    subtract_weights,
)


class _CubeWeightTester:
    # cube weight calculator
    cwc = None

    @pytest.mark.parametrize("inp", ["fail string", np.array([[1, 2], [3, 4]])])
    def test_get_land_mask_input_type_errors(self, test_all_cubes, inp):
        error_msg = re.escape(r"cube must be an ScmCube instance")
        masker = self.cwc(test_all_cubes)
        with pytest.raises(TypeError, match=error_msg):
            get_land_weights(masker, test_all_cubes, sftlf_cube=inp)

    def test_get_nh_mask(self, test_all_cubes):
        masker = self.cwc(test_all_cubes)
        result = get_nh_weights(masker, test_all_cubes)
        expected = np.array([[1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 0, 0]])

        np.testing.assert_array_equal(result, expected)

    def test_unknown_mask_error(self, test_all_cubes):
        masker = self.cwc(test_all_cubes)
        with pytest.raises(InvalidWeights, match="Unknown weights: junk"):
            masker.get_weights_array("junk")

    @patch(
        "netcdf_scm.weights.WEIGHTS_FUNCTIONS_WITHOUT_AREA_WEIGHTING",
        {
            "Junk": multiply_weights(get_weights_for_area(0, 0, 30, 50), "World|Land"),
            "World|Land": get_land_weights,
            "Inverse": subtract_weights("Junk", 1),
        },
    )
    def test_no_match_error(self, test_all_cubes):
        tmask_name = "Junk"

        error_msg = re.escape(
            r"All weights are zero for region: `{}`".format(tmask_name)
        )
        weighter = self.cwc(test_all_cubes)
        for i in range(3):  # make sure multiple asks still raises
            # should be accessible without issue
            weighter.get_weights_array("World|Land")
            with pytest.raises(ValueError, match=error_msg):
                weighter.get_weights_array("Junk")

            # should be able to get inverse without problem
            res = weighter.get_weights_array("Inverse")
            # inverse of Junk should all be non-zero
            assert not np.isclose(res, 0).any()


class TestAreaSurfaceFractionWeightCalculator(_CubeWeightTester):
    cwc = AreaSurfaceFractionWeightCalculator



import datetime as dt
import logging
import re

import pandas as pd
import pytest
import scipy.interpolate
from base_normalisation_integration import NormaliserIntegrationTester

from netcdf_scm.normalisation import NormaliserRunningMean
from netcdf_scm.stitching import get_branch_time


class TestRunningMean(NormaliserIntegrationTester):
    tclass = NormaliserRunningMean

    @pytest.mark.parametrize("member_id", ("r1i1p1f1", "r10i1p1f1"))
    @pytest.mark.parametrize("activity_id", ("ScenarioMIP", "RFMIP"))
    def test_get_reference_values(
        self,
        historical_data_normalisation,
        picontrol_data_normalisation,
        member_id,
        activity_id,
    ):
        # shorten the data so our picontrol data is long enough to be useable
        historical_data_to_use = historical_data_normalisation.filter(
            year=range(1850, 1950)
        ).copy()

        picontrol_branch_time = get_branch_time(historical_data_to_use)
        first_relevant_year = (
            picontrol_branch_time.year
            - 11  # 21-year running mean so go 11 years either side to ensure we have enough data
        )
        last_relevant_year = (
            picontrol_branch_time.year
            + historical_data_to_use["year"].max()
            - historical_data_to_use["year"].min()
            + 11  # 21-year running mean so go 11 years either side to ensure we have enough data
        )

        ref_ts = picontrol_data_normalisation.filter(
            year=range(first_relevant_year, last_relevant_year + 1)
        ).timeseries()
        ref_values = ref_ts.rolling(window=21 * 12, center=True, axis="columns").mean()
        relevant_cols = ref_values.columns[
            ref_values.columns.map(
                lambda x: (x.year >= picontrol_branch_time.year)
                and (x.year <= last_relevant_year - 11)
            )
        ]
        expected = ref_values[relevant_cols]

        expected = expected.reset_index()
        idx = historical_data_to_use.meta.columns.tolist()
        for c in idx:
            if c in ["region", "variable", "unit"]:
                continue

            expected[c] = historical_data_to_use.get_unique_meta(c, no_duplicates=True)

        expected = expected.set_index(idx)
        expected.columns = historical_data_to_use["time"]

        picontrol_data_to_use = picontrol_data_normalisation.copy()
        picontrol_data_to_use.set_meta(member_id, "member_id")
        picontrol_data_to_use.set_meta(activity_id, "activity_id")

        res = self.tclass().get_reference_values(
            historical_data_to_use, picontrol_data_to_use, picontrol_branch_time,
        )

        pd.testing.assert_frame_equal(res, expected, check_like=True)

    @pytest.mark.parametrize("member_id", ("r1i1p1f1", "r10i1p1f1"))
    @pytest.mark.parametrize("activity_id", ("ScenarioMIP", "RFMIP"))
    def test_get_reference_values_with_extension(
        self,
        historical_data_normalisation,
        picontrol_data_normalisation,
        member_id,
        activity_id,
        caplog,
    ):
        # shorten the data so our picontrol data is long enough to be useable
        historical_data_to_use = historical_data_normalisation.filter(
            year=range(1850, 2000)
        ).copy()

        picontrol_branch_time = get_branch_time(historical_data_to_use)
        first_relevant_year = (
            picontrol_branch_time.year
            - 5  # don't give quite enough data so some has to be filled
        )
        last_relevant_year = (
            picontrol_branch_time.year
            + historical_data_to_use["year"].max()
            - historical_data_to_use["year"].min()
            + 11  # 21-year running mean so go 11 years either side to ensure we have enough data
        )

        picontrol_data_to_use = picontrol_data_normalisation.filter(
            year=range(first_relevant_year, last_relevant_year + 1)
        ).copy()

        ref_ts = picontrol_data_to_use.timeseries()
        ref_values = ref_ts.rolling(window=21 * 12, center=True, axis="columns").mean()

        expected = ref_values.reset_index()
        idx = historical_data_to_use.meta.columns.tolist()
        for c in idx:
            if c in ["region", "variable", "unit"]:
                continue

            expected[c] = historical_data_to_use.get_unique_meta(c, no_duplicates=True)

        expected = expected.set_index(idx)

        relevant_cols = expected.columns[
            expected.columns.map(
                lambda x: (x.year >= picontrol_branch_time.year)
                and (x.year <= last_relevant_year - 11)
            )
        ]
        expected = expected[relevant_cols]

        interp_base = expected.dropna(axis=1)

        def _convert_to_s(x):
            return (x - dt.datetime(1970, 1, 1)).total_seconds()

        time_axis_base = interp_base.columns.map(_convert_to_s)
        interpolater = scipy.interpolate.interp1d(
            time_axis_base, interp_base.values, fill_value="extrapolate"
        )

        nan_cols = sorted(list(set(expected.columns) - set(interp_base.columns)))
        time_nan_cols = [_convert_to_s(x) for x in nan_cols]

        expected.loc[:, expected.isnull().any()] = interpolater(time_nan_cols)
        expected.columns = historical_data_to_use["time"]

        picontrol_data_to_use.set_meta(member_id, "member_id")
        picontrol_data_to_use.set_meta(activity_id, "activity_id")

        with caplog.at_level(logging.INFO):
            res = self.tclass().get_reference_values(
                historical_data_to_use, picontrol_data_to_use, picontrol_branch_time,
            )

        pd.testing.assert_frame_equal(res, expected, check_like=True)
        info_str = (
            "Filling gaps in running mean (where not enough values were available to create a full "
            "window) with linear extrapolations."
        )
        fill_info = [r for r in caplog.record_tuples if r[2] == info_str]
        assert len(fill_info) == 1
        assert fill_info[0][1] == logging.INFO

    @pytest.mark.parametrize(
        "start_cut,end_cut,expected_error",
        (
            (10 ** 3, -30, True),
            (10 ** 3, -20, True),
            (10 ** 3, -10, True),
            (10 ** 3, -5, True),
            (10 ** 3, -1, True),
            (10 ** 3, 0, False),
            (0, 10 ** 3, False),
            (10 ** 3, 5, False),
            (5, 10 ** 3, False),
        ),
    )
    def test_get_reference_values_not_enough_data_error(
        self,
        historical_data_normalisation,
        picontrol_data_normalisation,
        start_cut,
        end_cut,
        expected_error,
        caplog,
    ):
        # shorten the data so our picontrol data is long enough to be useable
        historical_data_to_use = historical_data_normalisation.filter(
            year=range(1850, 1950)
        ).copy()

        picontrol_branch_time = get_branch_time(historical_data_to_use)
        pi_data_to_use = picontrol_data_normalisation.filter(
            year=range(
                picontrol_branch_time.year - start_cut,
                picontrol_branch_time.year
                + historical_data_to_use["year"].max()
                - historical_data_to_use["year"].min()
                + end_cut
                + 1,
            )
        )

        error_msg = re.escape(
            "Only `{:04d}01` to `{:04d}12` is available in the piControl data. "
            "Given the branching time, `084101`, we need data from ~`0830` to `0951`. "
            "piControl data in "
            "{}".format(
                pi_data_to_use["year"].min(),
                pi_data_to_use["year"].max(),
                pi_data_to_use.metadata["netcdf-scm crunched file"],
            )
        )

        def call():
            self.tclass().get_reference_values(
                historical_data_to_use, pi_data_to_use, picontrol_branch_time,
            )

        if expected_error:
            with pytest.raises(ValueError, match=error_msg):
                call()

        else:
            with caplog.at_level(logging.INFO):
                call()

            info_str = (
                "Filling gaps in running mean (where not enough values were available to create a full "
                "window) with linear extrapolations."
            )
            info_msg = [r for r in caplog.record_tuples if r[2] == info_str]
            assert len(info_msg) == 1
            assert info_msg[0][1] == logging.INFO



"""
Parallelisation functions for the command-line operations

With some adaptation, the functions provided in this module could also be used
in other contexts. However, given they are in our 'private' API, we make no
guarantees about their usage.
"""
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed

import tqdm

try:
    import dask
except ModuleNotFoundError:  # pragma: no cover # emergency valve
    from .errors import raise_no_iris_warning

    raise_no_iris_warning()


def _apply_func(  # pylint:disable=too-many-arguments
    logger,
    apply_func,
    loop_kwarglist,
    common_arglist=None,
    common_kwarglist=None,
    postprocess_func=None,
    n_workers=2,
    style="processes",
    return_res=False,
):
    common_arglist = [] if common_arglist is None else common_arglist
    common_kwarglist = {} if common_kwarglist is None else common_kwarglist
    tqdm_kwargs = {
        "total": len(loop_kwarglist),
        "unit": "it",
        "unit_scale": True,
        "leave": True,
    }
    if n_workers == 1:
        failures = _apply_func_serially(
            logger=logger,
            apply_func=apply_func,
            loop_kwarglist=loop_kwarglist,
            tqdm_kwargs=tqdm_kwargs,
            common_arglist=common_arglist,
            common_kwarglist=common_kwarglist,
            postprocess_func=postprocess_func,
            return_res=return_res,
        )
    else:
        failures = _apply_func_parallel(
            logger=logger,
            apply_func=apply_func,
            loop_kwarglist=loop_kwarglist,
            tqdm_kwargs=tqdm_kwargs,
            common_arglist=common_arglist,
            common_kwarglist=common_kwarglist,
            postprocess_func=postprocess_func,
            n_workers=n_workers,
            style=style,
            return_res=return_res,
        )

    return failures


def _apply_func_serially(  # pylint:disable=too-many-arguments
    logger,
    apply_func,
    loop_kwarglist,
    tqdm_kwargs,
    common_arglist,
    common_kwarglist,
    postprocess_func,
    return_res=False,
):
    failures = False
    logger.info("Processing serially")
    full_res = []
    for ikwargs in tqdm.tqdm(loop_kwarglist, **tqdm_kwargs):
        try:
            res = apply_func(*common_arglist, **ikwargs, **common_kwarglist)
            if return_res:
                full_res.append(res)
            if postprocess_func is not None:
                postprocess_func(res)
        except Exception as e:  # pylint:disable=broad-except
            logger.exception("Exception found %s", e)
            failures = True

    if return_res:
        return full_res

    return failures


def _apply_func_parallel(  # pylint:disable=too-many-arguments
    logger,
    apply_func,
    loop_kwarglist,
    tqdm_kwargs,
    common_arglist,
    common_kwarglist,
    postprocess_func,
    n_workers,
    style,
    return_res=False,
):
    failures = False
    logger.info("Processing in parallel with %s workers", n_workers)
    logger.info("Forcing dask to use a single thread when reading")
    with dask.config.set(scheduler="single-threaded"):
        if style == "processes":
            executor_cls = ProcessPoolExecutor
        elif style == "threads":
            executor_cls = ThreadPoolExecutor
        else:
            raise ValueError("Unrecognised executor: {}".format(style))
        with executor_cls(max_workers=n_workers) as executor:
            futures = [
                executor.submit(
                    apply_func, *common_arglist, **ikwargs, **common_kwarglist
                )
                for ikwargs in loop_kwarglist
            ]
            failures = False
            full_res = []
            # Print out the progress as tasks complete
            for future in tqdm.tqdm(as_completed(futures), **tqdm_kwargs):
                try:
                    res = future.result()
                    if postprocess_func is not None:
                        postprocess_func(res)
                    if return_res:
                        full_res.append(res)
                except Exception as e:  # pylint:disable=broad-except
                    logger.exception("Exception found %s", e)
                    failures = True

    if return_res:
        return full_res

    return failures



import os
import re
import shutil

import numpy as np
import pytest

import netcdf_scm
from netcdf_scm.io import (
    _ALL_AVAILABLE_CUBES,
    get_scmcube_helper,
    load_mag_file,
    load_scmdataframe,
    save_netcdf_scm_nc,
)
from netcdf_scm.iris_cube_wrappers import (
    CMIP6Input4MIPsCube,
    CMIP6OutputCube,
    MarbleCMIP5Cube,
    ScmCube,
)


def _assert_scm_dataframe(scmdf, expected, **kwargs):
    d = scmdf.filter(**kwargs).timeseries()
    assert not d.empty
    np.testing.assert_allclose(d.values.squeeze(), expected)


def test_load_scmdataframe(test_data_netcdfscm_nc_file):
    loaded = load_scmdataframe(test_data_netcdfscm_nc_file)
    assert (loaded["scenario"] == "rcp45").all()
    assert (loaded["climate_model"] == "ACCESS1-0").all()
    assert (loaded["variable"] == "tas").all()
    assert (loaded["variable_standard_name"] == "air_temperature").all()
    assert (loaded["unit"] == "K").all()
    assert (loaded["member_id"] == "r1i1p1").all()
    assert (loaded["mip_era"] == "CMIP5").all()
    assert (loaded["activity_id"] == "cmip5").all()

    _assert_scm_dataframe(loaded, 285.521667, region="World", year=2006, month=1)
    _assert_scm_dataframe(loaded, 279.19043, region="World|Land", year=2019, month=3)
    _assert_scm_dataframe(
        loaded, 287.103729, region="World|Northern Hemisphere", year=2032, month=11
    )
    _assert_scm_dataframe(
        loaded,
        290.850189,
        region="World|Northern Hemisphere|Ocean",
        year=2049,
        month=12,
    )

    assert (
        loaded.metadata["institution"]
        == "CSIRO (Commonwealth Scientific and Industrial Research Organisation, Australia), and BOM (Bureau of Meteorology, Australia)"
    )
    assert (
        loaded.metadata["title"] == "ACCESS1-0 model output prepared for CMIP5 RCP4.5"
    )
    np.testing.assert_allclose(
        loaded.metadata["land_fraction_northern_hemisphere"], 0.38912639
    )


def test_save_cube_and_load_scmdataframe(tmpdir, test_cmip6_output_file):
    base = CMIP6OutputCube()
    base.load_data_from_path(test_cmip6_output_file)
    out_file = os.path.join(tmpdir, "test_save_file.nc")

    save_netcdf_scm_nc(base.get_scm_timeseries_cubes(), out_file)

    loaded = load_scmdataframe(out_file)
    assert (loaded["scenario"] == "1pctCO2").all()
    assert (loaded["climate_model"] == "BCC-CSM2-MR").all()
    assert (loaded["variable"] == "rlut").all()
    assert (loaded["variable_standard_name"] == "toa_outgoing_longwave_flux").all()
    assert (loaded["unit"] == "W m^-2").all()
    assert (loaded["activity_id"] == "CMIP").all()
    assert (loaded["member_id"] == "r1i1p1f1").all()
    assert (loaded["mip_era"] == "CMIP6").all()
    assert (loaded["activity_id"] == "CMIP").all()

    _assert_scm_dataframe(loaded, 236.569464, region="World", year=1859, month=12)
    _assert_scm_dataframe(loaded, 243.072575, region="World|Ocean", year=1856, month=10)
    _assert_scm_dataframe(
        loaded, 235.025871, region="World|Southern Hemisphere", year=1853, month=6
    )
    _assert_scm_dataframe(
        loaded, 234.333421, region="World|Southern Hemisphere|Land", year=1850, month=1
    )

    assert loaded.metadata[
        "crunch_netcdf_scm_version"
    ] == "{} (more info at gitlab.com/netcdf-scm/netcdf-scm)".format(
        netcdf_scm.__version__
    )
    assert (
        loaded.metadata["institution"]
        == "Beijing Climate Center, Beijing 100081, China"
    )
    assert loaded.metadata["title"] == "BCC-CSM2-MR output prepared for CMIP6"
    np.testing.assert_allclose(
        loaded.metadata["land_fraction_northern_hemisphere"], 0.38681179
    )
    assert (
        loaded.metadata["source"]
        == "BCC-CSM 2 MR (2017):   aerosol: none  atmos: BCC_AGCM3_MR (T106; 320 x 160 longitude/latitude; 46 levels; top level 1.46 hPa)  atmosChem: none  land: BCC_AVIM2  landIce: none  ocean: MOM4 (1/3 deg 10S-10N, 1/3-1 deg 10-30 N/S, and 1 deg in high latitudes; 360 x 232 longitude/latitude; 40 levels; top grid cell 0-10 m)  ocnBgchem: none  seaIce: SIS2"
    )


def test_all_available_cubes():
    expected = {}
    for c in dir(netcdf_scm.iris_cube_wrappers):
        try:
            tester = getattr(netcdf_scm.iris_cube_wrappers, c)
            if issubclass(tester, netcdf_scm.iris_cube_wrappers.ScmCube):
                if tester == netcdf_scm.iris_cube_wrappers._CMIPCube:
                    continue

                expected[c.replace("Cube", "")] = tester

        except TypeError:
            continue

    assert expected == _ALL_AVAILABLE_CUBES


@pytest.mark.parametrize(
    "drs,expected",
    (
        ("Scm", ScmCube),
        ("MarbleCMIP5", MarbleCMIP5Cube),
        ("CMIP6Input4MIPs", CMIP6Input4MIPsCube),
        ("CMIP6Output", CMIP6OutputCube),
    ),
)
def test_get_scmcube_helper(drs, expected):
    assert isinstance(get_scmcube_helper(drs), expected)


def test_get_scmcube_helper_none_error():
    error_msg = re.escape(
        "`drs` == 'None' is not supported yet. Please raise an issue at "
        "gitlab.com/netcdf-scm/netcdf-scm/ with your use case if you need this "
        "feature."
    )
    with pytest.raises(NotImplementedError, match=error_msg):
        get_scmcube_helper("None")


def test_get_scmcube_helper_unrecognised_error():
    error_msg = re.escape("Unrecognised drs: junk")
    with pytest.raises(KeyError, match=error_msg):
        get_scmcube_helper("junk")


@pytest.fixture
def test_mag_file_cmip6(test_cmip6_stitch_output):
    return os.path.join(
        test_cmip6_stitch_output,
        "mag-files-average-year-mid-year",
        "CMIP6",
        "ScenarioMIP",
        "BCC/BCC-CSM2-MR",
        "ssp126",
        "r1i1p1f1",
        "Amon",
        "tas",
        "gn",
        "v20190314",
        "netcdf-scm_tas_Amon_BCC-CSM2-MR_ssp126_r1i1p1f1_gn_1850-2100.MAG",
    )


@pytest.mark.parametrize("include_prefix", (True, False))
def test_load_mag_file_cmip6(test_mag_file_cmip6, include_prefix, tmpdir):
    if not include_prefix:
        load_name = os.path.join(
            tmpdir, test_mag_file_cmip6.replace("netcdf-scm_", "").strip(os.sep),
        )
        os.makedirs(os.path.dirname(load_name))
        shutil.copy(test_mag_file_cmip6, load_name)

    else:
        load_name = test_mag_file_cmip6

    res = load_mag_file(load_name, "CMIP6Output")

    assert res.get_unique_meta("scenario", no_duplicates=True) == "ssp126"
    assert res.get_unique_meta("climate_model", no_duplicates=True) == "BCC-CSM2-MR"
    assert res.get_unique_meta("activity_id", no_duplicates=True) == "ScenarioMIP"
    assert res.get_unique_meta("member_id", no_duplicates=True) == "r1i1p1f1"
    assert res.get_unique_meta("mip_era", no_duplicates=True) == "CMIP6"

    # # TODO: turn this on once we have an easy way to drop meta in scmdata
    # assert "todo" not in res.meta


@pytest.mark.parametrize("include_prefix", (True, False))
def test_load_mag_file_cmip6_name_only(test_mag_file_cmip6, include_prefix, tmpdir):
    load_name = os.path.basename(test_mag_file_cmip6)
    if not include_prefix:
        load_name = load_name.replace("netcdf-scm_", "")

    load_name = os.path.join(tmpdir, load_name)
    shutil.copy(test_mag_file_cmip6, load_name)

    with pytest.warns(UserWarning) as record:
        res = load_mag_file(load_name, "CMIP6Output")

    assert len(record) == 4
    assert (
        record[1]
        .message.args[0]
        .startswith(
            "Couldn't determine cube metadata from path with the drs 'CMIP6Output'."
            "\nError:"
        )
    )
    assert record[2].message.args[0] == (
        "Can't set metadata for 'activity_id' from filename alone with the "
        "drs 'CMIP6Output'."
    )
    assert record[3].message.args[0] == (
        "Can't set metadata for 'mip_era' from filename alone with the drs "
        "'CMIP6Output'."
    )

    assert res.get_unique_meta("scenario", no_duplicates=True) == "ssp126"
    assert res.get_unique_meta("climate_model", no_duplicates=True) == "BCC-CSM2-MR"
    assert res.get_unique_meta("member_id", no_duplicates=True) == "r1i1p1f1"

    # can't be identified from filename alone
    assert "activity_id" not in res.meta
    assert "mip_era" not in res.meta

    # # TODO: turn this on once we have an easy way to drop meta in scmdata
    # assert "todo" not in res.meta


def test_load_mag_file_cmip6_garbled_name(test_mag_file_cmip6, tmpdir):
    load_name = os.path.join(tmpdir, "GARBLED.MAG")
    shutil.copy(test_mag_file_cmip6, load_name)

    with pytest.warns(UserWarning) as record:
        res = load_mag_file(load_name, "CMIP6Output")

    assert len(record) == 3
    assert (
        record[1]
        .message.args[0]
        .startswith(
            "Couldn't determine cube metadata from path with the drs 'CMIP6Output'.\n"
            "Error:"
        )
    )
    assert (
        record[2]
        .message.args[0]
        .startswith(
            "Couldn't determine cube metadata from filename either with the drs "
            "'CMIP6Output'.\nError:"
        )
    )

    assert res.get_unique_meta("scenario", no_duplicates=True) == "unspecified"
    assert res.get_unique_meta("climate_model", no_duplicates=True) == "unspecified"

    # can't be identified hence not included
    assert "activity_id" not in res.meta
    assert "mip_era" not in res.meta
    assert "member_id" not in res.meta

    # # TODO: turn this on once we have an easy way to drop meta in scmdata
    # assert "todo" not in res.meta


@pytest.fixture
def test_mag_file_marble_cmip5(test_marble_cmip5_stitch_output):
    return os.path.join(
        test_marble_cmip5_stitch_output,
        "normalised",
        "21-yr-running-mean",
        "mag-files",
        "cmip5",
        "rcp45",
        "Amon",
        "tas",
        "NorESM1-M",
        "r1i1p1",
        "netcdf-scm_tas_Amon_NorESM1-M_rcp45_r1i1p1_185001-230012.MAG",
    )


@pytest.mark.parametrize("include_prefix", (True, False))
def test_load_mag_file_cmip5(test_mag_file_marble_cmip5, include_prefix, tmpdir):
    if not include_prefix:
        load_name = os.path.join(
            tmpdir, test_mag_file_marble_cmip5.replace("netcdf-scm_", "").strip(os.sep),
        )
        os.makedirs(os.path.dirname(load_name))
        shutil.copy(test_mag_file_marble_cmip5, load_name)

    else:
        load_name = test_mag_file_marble_cmip5

    res = load_mag_file(load_name, "MarbleCMIP5")

    assert res.get_unique_meta("scenario", no_duplicates=True) == "rcp45"
    assert res.get_unique_meta("climate_model", no_duplicates=True) == "NorESM1-M"
    assert res.get_unique_meta("member_id", no_duplicates=True) == "r1i1p1"
    assert res.get_unique_meta("mip_era", no_duplicates=True) == "CMIP5"
    assert res.get_unique_meta("activity_id", no_duplicates=True) == "cmip5"

    # # TODO: turn this on once we have an easy way to drop meta in scmdata
    # assert "todo" not in res.meta


@pytest.mark.parametrize("include_prefix", (True, False))
def test_load_mag_file_cmip5_name_only(
    test_mag_file_marble_cmip5, include_prefix, tmpdir
):
    load_name = os.path.basename(test_mag_file_marble_cmip5)
    if not include_prefix:
        load_name = load_name.replace("netcdf-scm_", "")

    load_name = os.path.join(tmpdir, load_name)
    shutil.copy(test_mag_file_marble_cmip5, load_name)

    with pytest.warns(UserWarning) as record:
        res = load_mag_file(load_name, "MarbleCMIP5")

    assert len(record) == 4
    assert (
        record[1]
        .message.args[0]
        .startswith(
            "Couldn't determine cube metadata from path with the drs 'MarbleCMIP5'."
            "\nError:"
        )
    )
    assert record[2].message.args[0] == (
        "Can't set metadata for 'activity_id' from filename alone with the "
        "drs 'MarbleCMIP5'."
    )
    assert record[3].message.args[0] == (
        "Can't set metadata for 'mip_era' from filename alone with the drs "
        "'MarbleCMIP5'."
    )

    assert res.get_unique_meta("scenario", no_duplicates=True) == "rcp45"
    assert res.get_unique_meta("climate_model", no_duplicates=True) == "NorESM1-M"
    assert res.get_unique_meta("member_id", no_duplicates=True) == "r1i1p1"

    # can't be identified from filename alone
    assert "activity_id" not in res.meta
    assert "mip_era" not in res.meta

    # # TODO: turn this on once we have an easy way to drop meta in scmdata
    # assert "todo" not in res.meta


def test_load_mag_file_cmip5_garbled_name(test_mag_file_marble_cmip5, tmpdir):
    load_name = os.path.join(tmpdir, "GARBLED.MAG")
    shutil.copy(test_mag_file_marble_cmip5, load_name)

    with pytest.warns(UserWarning) as record:
        res = load_mag_file(load_name, "MarbleCMIP5")

    assert len(record) == 3
    assert (
        record[1]
        .message.args[0]
        .startswith(
            "Couldn't determine cube metadata from path with the drs 'MarbleCMIP5'.\n"
            "Error:"
        )
    )
    assert (
        record[2]
        .message.args[0]
        .startswith(
            "Couldn't determine cube metadata from filename either with the drs "
            "'MarbleCMIP5'.\nError:"
        )
    )

    assert res.get_unique_meta("scenario", no_duplicates=True) == "unspecified"
    assert res.get_unique_meta("climate_model", no_duplicates=True) == "unspecified"

    # can't be identified hence not included
    assert "activity_id" not in res.meta
    assert "mip_era" not in res.meta
    assert "member_id" not in res.meta

    # # TODO: turn this on once we have an easy way to drop meta in scmdata
    # assert "todo" not in res.meta



from base_normalisation_unit import NormaliserTester

from netcdf_scm.normalisation import NormaliserRunningMean


class TestRunningMean(NormaliserTester):
    tclass = NormaliserRunningMean

    def test_method_name(self):
        assert self.tclass().method_name == "21-yr-running-mean"



"""
Module for the normaliser which calculates anomalies from a running mean in the pre-industrial control run
"""

import datetime as dt
import logging

import scipy.interpolate as scipyinterp

from ..cli_utils import _check_timesteps_are_monthly
from .base import Normaliser

LOGGER = logging.getLogger(__name__)


class NormaliserRunningMean(Normaliser):
    """
    Normaliser which calculates anomalies from a running mean in the pre-industrial control run

    Each normalisation value is a 21-year mean, centred on the equivalent point in the
    pre-industrial control simulation. If there is insufficient data to create a full 21-year
    window at the edge of the simulation then a linear extrapolation of the running-mean is used
    to extend the normalisation values to cover the required full range.

    At present, only a 21-year running mean is implemented.
    """

    def __init__(self):
        self._method_name = "21-yr-running-mean"

    @staticmethod
    def _get_reference_values(indata, picontrol, picontrol_branching_time):
        running_mean_n_years = 21
        months_per_year = 12
        window_width = running_mean_n_years * months_per_year
        window_roll_back = window_width // 2
        # the window extends one step fewer forward because the centre time point
        # is also included
        window_roll_forward = window_roll_back - 1

        branch_year = picontrol_branching_time.year
        branch_month = picontrol_branching_time.month
        picontrol_branch_col = picontrol.filter(year=branch_year, month=branch_month)[
            "time"
        ][0]
        picontrol_time = picontrol["time"]
        picontrol_branch_idx = (picontrol_time == picontrol_branch_col).idxmax()

        # relevant times go from the branch point minus the number of steps each
        # rolling mean window extends backward up to the branch point plus the
        # length of ``indata`` plus the number of steps the rolling mean window
        # extends forward
        relevant_time_points = picontrol_time[
            max(picontrol_branch_idx - window_roll_back, 0) : picontrol_branch_idx
            + indata["time"].shape[0]
            + window_roll_forward
        ].tolist()

        _check_timesteps_are_monthly(picontrol)  # quick sanity check

        # super slow option below, would be good to have direct functionality in scmdata
        # normalise_ts = picontrol.filter(time=relevant_time_points.tolist()).timeseries()

        picontrol_times = (picontrol["time"] >= relevant_time_points[0]) & (
            picontrol["time"] <= relevant_time_points[-1]
        )
        normalise_ts = picontrol.timeseries().iloc[:, picontrol_times.values]

        normalise_df = normalise_ts.rolling(
            window=window_width, center=True, axis="columns",
        ).mean()

        keep_time_points = picontrol_time[
            picontrol_branch_idx : picontrol_branch_idx + indata["time"].shape[0]
        ].tolist()

        raise_error = False
        try:
            normalise_df = normalise_df[keep_time_points]
        except KeyError:
            raise_error = True

        if normalise_df.isnull().any().any():
            LOGGER.info(
                "Filling gaps in running mean (where not enough values were available to create a full "
                "window) with linear extrapolations."
            )
            # # TODO: fix, this should work but is broken...
            # ScmDataFrame(normalise_df).interpolate(target_times=normalise_df.columns, extrapolation_type="linear").timeseries()

            interp_base = normalise_df.dropna(axis=1)

            def _convert_to_s(x):
                return (x - dt.datetime(1970, 1, 1)).total_seconds()

            time_axis_base = interp_base.columns.map(_convert_to_s)
            interpolater = scipyinterp.interp1d(
                time_axis_base, interp_base.values, fill_value="extrapolate"
            )

            nan_cols = sorted(
                list(set(normalise_df.columns) - set(interp_base.columns))
            )
            time_nan_cols = [_convert_to_s(x) for x in nan_cols]

            normalise_df.loc[:, normalise_df.isnull().any()] = interpolater(
                time_nan_cols
            )

        if normalise_df.isnull().any().any():
            # if there are any remaining gaps, raise
            raise_error = True

        try:
            normalise_df.columns = indata["time"]
        except ValueError:
            raise_error = True

        if raise_error:
            error_msg = (
                "Only `{:04d}{:02d}` to `{:04d}{:02d}` is available in the "
                "{} data. Given the branching time, `{:04d}{:02d}`, we need "
                "data from ~`{:04d}` to `{:04d}`. {} data in {}".format(
                    picontrol["time"].min().year,
                    picontrol["time"].min().month,
                    picontrol["time"].max().year,
                    picontrol["time"].max().month,
                    picontrol.metadata["experiment_id"],
                    picontrol_branching_time.year,
                    picontrol_branching_time.month,
                    picontrol_branching_time.year - 11,
                    picontrol_branching_time.year
                    + indata["year"].max()
                    - indata["year"].min()
                    + 11,
                    picontrol.metadata["experiment_id"],
                    picontrol.metadata["netcdf-scm crunched file"],
                )
            )
            raise ValueError(error_msg)

        return normalise_df



"""
Module for the normaliser which calculates anomalies from a mean of a fixed number of years in the pre-industrial control run
"""

import logging

import pandas as pd

from .base import Normaliser

LOGGER = logging.getLogger(__name__)


class AfterBranchTimeMean(Normaliser):
    """
    Normaliser which calculates anomalies from a mean of a fixed number of years after the branch time in the pre-industrial control run

    At present, only a 31-year mean after the branch time is implemented.
    """

    def __init__(self):
        self._method_name = "31-yr-mean-after-branch-time"

    @staticmethod
    def _get_reference_values(indata, picontrol, picontrol_branching_time):
        branch_year = picontrol_branching_time.year
        normalise_series = picontrol.filter(year=range(branch_year, branch_year + 31))

        if (
            normalise_series["time"].max().year - normalise_series["time"].min().year
        ) != 30:
            error_msg = (
                "Only `{:04d}{:02d}` to `{:04d}{:02d}` is available after the "
                "branching time `{:04d}{:02d}` in {} data in {}".format(
                    normalise_series["time"].min().year,
                    normalise_series["time"].min().month,
                    normalise_series["time"].max().year,
                    normalise_series["time"].max().month,
                    picontrol_branching_time.year,
                    picontrol_branching_time.month,
                    picontrol.metadata["experiment_id"],
                    picontrol.metadata["netcdf-scm crunched file"],
                )
            )
            raise ValueError(error_msg)

        normalise_mean = normalise_series.timeseries().mean(axis=1)
        # normalisation is uniform for all timepoints
        normalise_df = pd.concat([normalise_mean] * indata["time"].shape[0], axis=1)
        normalise_df.columns = indata["time"]

        return normalise_df



import datetime as dt
import re
from os.path import join

import numpy as np
import pandas as pd
import pytest
from conftest import TEST_DATA_ROOT_DIR
from scmdata import ScmDataFrame

from netcdf_scm.wranglers import (
    convert_scmdf_to_tuningstruc,
    convert_tuningstruc_to_scmdf,
    get_tuningstruc_name_from_df,
)

TEST_DATA_TUNINGSTRUCS_DIR = join(TEST_DATA_ROOT_DIR, "tuningstrucs")


@pytest.fixture(
    scope="function",
    params=[
        {
            "location": join(
                TEST_DATA_TUNINGSTRUCS_DIR,
                "xavier_RCP85_GISSLE_transient_SMBdata_1850nan.mat",
            ),
            "var": "SMB",
            "region": "World",
            "unit": "mm",
            "scenario": "RCP85",
            "climate_model_2100_values": {
                "CCSM4": 47.64728531855956,
                "FGOALSG2": 92.01503047091413,
                "GISSE2R": 56.65268144044319,
                "IPSLCM5ALR": 124.97478670360111,
            },
        },
        {
            "location": join(TEST_DATA_TUNINGSTRUCS_DIR, "single_var_tuningstruc.mat"),
            "var": "cLitter",
            "region": "World",
            "unit": "GtC",
            "scenario": "RCP26",
            "climate_model_2100_values": {"BNUESM": 0.30397183},
        },
        {
            "location": join(TEST_DATA_TUNINGSTRUCS_DIR, "CABLE_rcp85_C0_npp.mat"),
            "var": "npp",
            "region": "World",
            "unit": "PgC/yr",
            "scenario": "RCP85",
            "climate_model_2100_values": {"CABLE": 30.967642},
        },
    ],
)
def test_file_info(request):
    yield request.param


@pytest.mark.parametrize("model", [None, "junk"])
def test_convert_tuningstruc_to_scmdf(test_file_info, model):
    test_file = test_file_info["location"]
    tkwargs = {"model": model}
    tvar = test_file_info["var"]
    tregion = test_file_info["region"]
    tunit = test_file_info["unit"]
    tscen = test_file_info["scenario"]

    res = convert_tuningstruc_to_scmdf(
        test_file, tvar, tregion, tunit, tscen, **tkwargs
    )

    if model is None:
        assert (res["model"] == "unspecified").all()  # default
    else:
        assert (res["model"] == model).all()

    for cm, v in test_file_info["climate_model_2100_values"].items():
        rv = (
            res.filter(
                climate_model=cm,
                year=2100,
                variable=tvar,
                region=tregion,
                scenario=tscen,
                unit=tunit,
            )
            .timeseries()
            .values
        )
        np.testing.assert_allclose(rv, v)


def test_convert_scmdf_to_tuningstruc_single_char_unit(tmpdir):
    test_df = ScmDataFrame(
        np.array([1, 2, 3]),
        index=[dt.datetime(y, 1, 1) for y in [1990, 1991, 1992]],
        columns={
            "variable": "var",
            "region": "World",
            "unit": "K",
            "scenario": "test-scenario",
            "model": "test_model",
            "climate_model": "test_cm",
            "member_id": "tmember-id",
        },
    )

    convert_scmdf_to_tuningstruc(test_df, tmpdir, prefix="test_tuningstruc")
    expected_outfile = join(
        tmpdir, "test_tuningstruc_VAR_TEST-SCENARIO_TMEMBER-ID_WORLD.mat"
    )

    reread = convert_tuningstruc_to_scmdf(expected_outfile)
    assert (reread["unit"] == "K").all()


def test_convert_tuningstruc_to_scmdf_errors(test_file_info):
    test_file = test_file_info["location"]
    error_msg = r"Cannot determine \S* " + re.escape("from file: {}".format(test_file))
    with pytest.raises(KeyError, match=error_msg):
        convert_tuningstruc_to_scmdf(test_file)


def test_convert_scmdf_to_tuningstruc(test_file_info, tmpdir):
    test_file = test_file_info["location"]
    tvar = test_file_info["var"]
    tregion = "World|Northern Hemisphere|Ocean"
    tunit = test_file_info["unit"]
    tscen = test_file_info["scenario"]
    tmodel = "iam"
    start = convert_tuningstruc_to_scmdf(
        test_file, tvar, tregion, tunit, tscen, model=tmodel
    )

    expected_outfile = (
        join(
            tmpdir,
            "{}_{}_MEMBER-ID_{}".format(tvar, tscen, tregion)
            .replace(" ", "_")
            .replace("|", "_")
            .upper(),
        )
        + ".mat"
    )
    convert_scmdf_to_tuningstruc(start, tmpdir)
    res = convert_tuningstruc_to_scmdf(expected_outfile)

    pd.testing.assert_frame_equal(start.timeseries(), res.timeseries())


def test_get_tuningstruc_name_from_df_error():
    df = pd.DataFrame(["scen_1", "scen_2"], columns=["scenario"])
    with pytest.raises(ValueError, match="More than one scenario in df"):
        get_tuningstruc_name_from_df(df, "test", "filler")



"""
Utilities for checking for retracted datasets
"""
import concurrent.futures
import glob
import logging
import os.path
from collections import defaultdict
from urllib.parse import urlencode

import pandas as pd
import pymagicc.io
import requests
import tqdm.autonotebook as tqdman

from .iris_cube_wrappers import CMIP6OutputCube

LOGGER = logging.getLogger(__name__)

ESGF_NODE = "esgf.nci.org.au"


def _create_search_url(**kwargs):
    kwargs.setdefault("format", "application/solr+json")
    kwargs.setdefault("offset", 0)
    kwargs.setdefault("limit", 100)
    kwargs.setdefault("type", "Dataset")
    kwargs.setdefault("replica", False)

    return "https://{}/esg-search/search?{}".format(
        ESGF_NODE, urlencode(kwargs, doseq=True)
    )


def _query_esgf(instance_ids):
    response = requests.get(_create_search_url(instance_id=instance_ids))
    response.raise_for_status()
    response_json = response.json()

    return response_json, instance_ids


def _perform_esgf_queries(instance_ids, esgf_query_batch_size, nworkers):
    futures = []
    with concurrent.futures.ThreadPoolExecutor(max_workers=nworkers) as executor:
        for i in tqdman.tqdm(
            range(0, len(instance_ids), esgf_query_batch_size),
            desc="Querying ESGF (submitting jobs)",
        ):
            batch = instance_ids[i : i + esgf_query_batch_size]
            futures.append(executor.submit(_query_esgf, batch,))

        return [
            f.result()
            for f in tqdman.tqdm(
                concurrent.futures.as_completed(futures),
                total=len(futures),
                desc="Retrieving results from ESGF jobs",
            )
        ]


def check_retractions(instance_ids, esgf_query_batch_size=100, nworkers=8):
    """
    Check a list of ``instance_ids`` for any retracted datasets

    Notes
    -----
    This queries external ESGF servers. Please limit the number of parallel requests.

    Parameters
    ----------
    instance_ids : list of str
        Datasets to check. ``instance_id`` is the unique identifier for a dataset,
        for example `CMIP6.CMIP.CSIRO.ACCESS-ESM1-5.esm-hist.r1i1p1f1.Amon.rsut.gn.v20191128`

    esgf_query_batch_size : int
        Maximum number of ids to include in each query.

    nworkers : int
        Number of workers to parallel queries to ESGF.

    Returns
    -------
    list of str
        A list of retracted ``instance_ids``
    """
    response_jsons_search_instance_ids = _perform_esgf_queries(
        instance_ids, esgf_query_batch_size, nworkers
    )

    retracted_files = []
    for response_json, search_ids in response_jsons_search_instance_ids:
        found_ids = []
        for data_doc in response_json["response"]["docs"]:
            id_here = data_doc["instance_id"]
            found_ids.append(id_here)

            if data_doc["retracted"]:
                LOGGER.info("Processing dependents of retracted dataset: %s", id_here)
                retracted_files.append(id_here)

        if len(found_ids) != len(search_ids):
            missing_ids = set(search_ids) - set(found_ids)
            for mid in missing_ids:
                LOGGER.warning(
                    "Couldn't find %s (search url: %s)",
                    mid,
                    _create_search_url(instance_id=mid),
                )

    return retracted_files


def check_retracted_files(filenames_or_dir, filename_filter="*.nc", **kwargs):
    """
    Check if any files are retracted

    Notes
    -----
    This queries external ESGF servers. Please limit the number of parallel requests.

    Parameters
    ----------
    filenames_or_dir: list of str or str
        A list of filenames or a directory to check for any retractions. If a string
        is provided, it is assumed to reference a directory and any files within
        that directory matching the `filename_filter` will be checked.

    filename_filter: str
        If a directory is passed all files matching the filter will be checked.

    **kwargs : any
        Passed to :func:`check_retracted`

    Returns
    -------
    List of the retracted files
    """

    if isinstance(filenames_or_dir, str):
        if os.path.isdir(filenames_or_dir):
            fnames = glob.glob(
                os.path.join(filenames_or_dir, "**", filename_filter), recursive=True,
            )
        else:
            raise NotADirectoryError(filenames_or_dir)
    else:
        if not isinstance(filenames_or_dir, list):
            raise ValueError("filenames_or_dir should be a list of filenames")
        fnames = filenames_or_dir

    filepaths_instance_ids = []
    for fname in tqdman.tqdm(fnames):
        try:
            filepaths_instance_ids.append(
                (fname, CMIP6OutputCube.get_instance_id(fname))
            )
        except ValueError:
            LOGGER.exception("could not get instance_id from {}".format(fname))

    instance_ids = list(set([v[1] for v in filepaths_instance_ids]))

    retracted = check_retractions(instance_ids, **kwargs)

    retracted_files = []
    for fp, instance_id in filepaths_instance_ids:
        if instance_id in retracted:
            retracted_files.append(fp)

    return retracted_files


def check_depends_on_retracted(mag_files, raise_on_mismatch=True, **kwargs):
    """
    Check if a ``.MAG`` file was calculated from now retracted data

    Notes
    -----
    This queries external ESGF servers. Please limit the number of parallel requests.

    Parameters
    ----------
    mag_files : list of str
        List of ``.MAG`` files to check

    raise_on_mismatch : bool
        If a file cannot be processed, should an error be raised? If ``False``, an
        error message is logged instead.

    **kwargs : any
        Passed to :func:`check_retractions`

    Returns
    -------
    pd.DataFrame
        Dataframe which describes the retracted status of each file in ``mag_files``. The columns are:

            - "mag_file": the files in ``mag_files``

            - "dependency_file": file which the file in the "mag_file" column depends on (note that
                                 the ``.MAG`` files may have more than one dependency so they may
                                 appear more than once in the "mag_file" column)

            - "dependency_instance_id": instance id (i.e. unique ESGF identifier) of the dependency file

            - "dependency_retracted": whether the dependency file has been retracted or not (``True`` if
                                      the file has been retracated)

        The list of retracted ``.MAG`` files can then be accessed with e.g.
        ``res.loc[res["dependency_retracted"], "mag_file"].unique()``

    Raises
    ------
    ValueError
        The ``.MAG`` file is not based on CMIP6 data (retractions cannot be checked
        automatically for CMIP5 data with netCDF-SCM).

    ValueError
        Metadata about a ``.MAG`` file's source is not included in the ``.MAG`` file.
    """
    no_source_data_files = []
    non_cmip6_data_files = []

    out_table = defaultdict(list)

    for mf in tqdman.tqdm(mag_files):
        LOGGER.debug("Processing %s", mf)

        metadata = pymagicc.io.read_mag_file_metadata(mf)
        dependency_files = _get_dependencies(metadata)

        if not dependency_files:
            no_source_data_files.append(mf)
            continue

        for df in dependency_files:
            try:
                instance_id = CMIP6OutputCube.get_instance_id(df)
            except ValueError:
                non_cmip6_data_files.append(mf)
                continue

            out_table["mag_file"].append(mf)
            out_table["dependency_file"].append(df)
            out_table["dependency_instance_id"].append(instance_id)

    if no_source_data_files:
        msg = "Could not determine source data for: {}".format(
            set(no_source_data_files)
        )
        if raise_on_mismatch:
            raise ValueError(msg)

        LOGGER.error(msg)

    if non_cmip6_data_files:
        msg = "Source data is not CMIP6 for: {}".format(set(non_cmip6_data_files))
        if raise_on_mismatch:
            raise ValueError(msg)

        LOGGER.error(msg)

    out_table = pd.DataFrame(out_table)

    instance_ids_retracted = check_retractions(
        out_table["dependency_instance_id"].unique().tolist(), **kwargs
    )

    out_table["dependency_retracted"] = out_table["dependency_instance_id"].isin(
        instance_ids_retracted
    )

    return out_table


def _get_dependencies(metadata):
    dependency_files = []

    for k in metadata:
        if "source_files" in k:
            dependency_files_key = [
                v.strip("]").strip("'")
                for g in metadata[k].split(";")
                for sg in g.split(", ")
                for v in sg.split("[")[1:]
            ]

            if any(["," in d for d in dependency_files_key]):
                raise ValueError(
                    "Bad processing of {}, got {}".format(
                        metadata, dependency_files_key
                    )
                )

            dependency_files += dependency_files_key

    LOGGER.debug("Dependent files %s", dependency_files)

    return dependency_files



from os.path import join

import pytest
from click.testing import CliRunner

from netcdf_scm.cli import stitch_netcdf_scm_ncs


@pytest.mark.parametrize(
    "out_format", (["mag-files", "mag-files-average-year-mid-year"])
)
def test_stitching_results(
    tmpdir,
    update_expected_files,
    test_cmip6_crunch_output,
    test_cmip6_stitch_output,
    run_wrangling_comparison,
    out_format,
):
    input_dir = test_cmip6_crunch_output
    output_dir = str(tmpdir)

    runner = CliRunner()
    result = runner.invoke(
        stitch_netcdf_scm_ncs,
        [
            input_dir,
            output_dir,
            "cmip6output wrangling regression test",
            "--drs",
            "CMIP6Output",
            "-f",
            "--number-workers",
            1,
            "--out-format",
            out_format,
            # have to avoid files which will fail stiching and averaging
            "--regexp",
            "^(?!.*(piControl|thetao|hist-stratO3|CNRM-ESM2-1/ssp534-over|IPSL-CM6A-LR/.*/r1i1p1f1/Lmon/cSoilFast|IPSL-CM6A-LR/.*/r1i1p1f1/Lmon/gpp|CESM2|IPSL-CM6A-LR|CanESM5/.*hfds)).*$",
        ],
    )
    assert result.exit_code == 0, result.output

    run_wrangling_comparison(
        join(output_dir, "CMIP6"),
        join(test_cmip6_stitch_output, out_format, "CMIP6"),
        update=update_expected_files,
    )


@pytest.mark.parametrize(
    "out_format",
    (
        [
            "mag-files",
            "magicc-input-files-average-year-start-year",
            "magicc-input-files-point-start-year",
        ]
    ),
)
def test_stitching_and_normalise_results(
    tmpdir,
    update_expected_files,
    test_cmip6_crunch_output,
    test_cmip6_stitch_output,
    run_wrangling_comparison,
    out_format,
    normalisation_method,
):
    if normalisation_method.startswith("21-yr-running-mean"):
        pytest.xfail(
            "Not enough piControl data to do 21 year running mean normalisation"
        )

    input_dir = test_cmip6_crunch_output
    output_dir = join(str(tmpdir), normalisation_method)

    runner = CliRunner()
    result = runner.invoke(
        stitch_netcdf_scm_ncs,
        [
            input_dir,
            output_dir,
            "cmip6output stitching with {} normalisation regression test".format(
                normalisation_method
            ),
            "--drs",
            "CMIP6Output",
            "-f",
            "--number-workers",
            1,
            "--out-format",
            out_format,
            # have to avoid files which will fail normalising
            "--regexp",
            "^(?!.*(piControl|CNRM-CM6-1/hist-aer|GFDL-CM4/1pctCO2|GFDL-CM4/abrupt-4xCO2|CESM2/historical|IPSL-CM6A-LR/historical|GISS-E2-1-G/abrupt-4xCO2)).*/tas/.*$",
            "--normalise",
            normalisation_method,
        ],
    )
    assert result.exit_code == 0, result.output

    run_wrangling_comparison(
        join(output_dir, "CMIP6"),
        join(
            test_cmip6_stitch_output,
            "normalised",
            normalisation_method,
            out_format,
            "CMIP6",
        ),
        update=update_expected_files,
    )



import datetime as dt
import os
from glob import glob
from os.path import isdir, join

import numpy as np
import pandas as pd
import pytest
from click.testing import CliRunner
from pymagicc.io import MAGICCData

import netcdf_scm
from netcdf_scm.cli import wrangle_netcdf_scm_ncs
from netcdf_scm.iris_cube_wrappers import ScmCube


def test_wrangling_unsupported_format(tmpdir, caplog, test_cmip6_crunch_output):
    input_dir = test_cmip6_crunch_output
    output_dir = str(join(tmpdir, "new-sub-dir"))

    runner = CliRunner()
    with caplog.at_level("INFO"):
        result = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [input_dir, output_dir, "test-invalid-format", "--out-format", "junk"],
        )
    assert result.exit_code != 0
    assert "invalid choice: junk" in result.output
    assert isinstance(result.exception, SystemExit)


def test_wrangling_defaults(tmpdir, caplog, test_cmip6_crunch_output):
    input_dir = test_cmip6_crunch_output
    output_dir = str(join(tmpdir, "new-sub-dir"))
    test_wrangler = "test-defaults"

    runner = CliRunner()
    with caplog.at_level("INFO"):
        result = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [
                input_dir,
                output_dir,
                test_wrangler,
                "--drs",
                "CMIP6Output",
                "--number-workers",
                1,
                # avoid thetao which we can't wrangle yet
                "--regexp",
                "^((?!thetao).)*$",
            ],
        )
    assert result.exit_code == 0

    assert "netcdf-scm: {}".format(netcdf_scm.__version__) in result.output
    assert "Making output directory: {}".format(output_dir) in result.output

    assert "rlut" in result.output
    assert "lai" in result.output
    assert "cSoilFast" in result.output

    assert isdir(
        join(
            output_dir,
            "CMIP6/CMIP/BCC/BCC-CSM2-MR/1pctCO2/r1i1p1f1/Amon/rlut/gn/v20181015",
        )
    )
    assert isdir(
        join(
            output_dir,
            "CMIP6/ScenarioMIP/IPSL/IPSL-CM6A-LR/ssp126/r1i1p1f1/Lmon/cSoilFast/gr/v20190121",
        )
    )
    assert isdir(join(output_dir, "CMIP6/CMIP/CNRM-CERFACS"))

    with open(
        join(
            output_dir,
            "CMIP6/CMIP/CNRM-CERFACS/CNRM-CM6-1/historical/r1i1p1f2/Lmon/lai/gr/v20180917/netcdf-scm_lai_Lmon_CNRM-CM6-1_historical_r1i1p1f2_gr_200201-200512.MAG",
        )
    ) as f:
        content = f.read()

    assert "Contact: {}".format(test_wrangler) in content

    assert "Duplicating" not in result.output


def test_wrangling_magicc_input_files(tmpdir, caplog, test_marble_cmip5_crunch_output):
    input_dir = test_marble_cmip5_crunch_output
    output_dir = str(join(tmpdir, "new-sub-dir"))
    test_wrangler = "test wrangling magicc input point end of year files <email>"
    runner = CliRunner()
    with caplog.at_level("INFO"):
        result = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [
                input_dir,
                output_dir,
                test_wrangler,
                "--out-format",
                "magicc-input-files-point-end-year",
                "--regexp",
                "^((?!historical).)*tas.*$",
                "--drs",
                "MarbleCMIP5",
                "--number-workers",
                1,
            ],
        )
    assert result.exit_code == 0

    assert "netcdf-scm: {}".format(netcdf_scm.__version__) in result.output
    assert "Making output directory: {}".format(output_dir) in result.output

    assert isdir(join(output_dir, "cmip5/1pctCO2/Amon/tas/CanESM2/r1i1p1"))
    assert isdir(join(output_dir, "cmip5/rcp26/Amon/tas/bcc-csm1-1/r1i1p1"))
    assert isdir(join(output_dir, "cmip5/rcp45/Amon/tas/HadCM3/r1i1p1"))
    assert isdir(join(output_dir, "cmip5/rcp45/Amon/tas/ACCESS1-0/r1i1p1"))
    assert isdir(join(output_dir, "cmip5/rcp85/Amon/tas/NorESM1-ME/r1i1p1"))

    with open(
        join(
            output_dir,
            "cmip5/rcp45/Amon/tas/HadCM3/r1i1p1/TAS_RCP45_HADCM3_R1I1P1_GLOBAL_SURFACE_TEMP.IN",
        )
    ) as f:
        content = f.read()

    assert "Contact: {}".format(test_wrangler) in content
    assert "timeseriestype: POINT_END_YEAR" in content
    assert "2006" in content
    assert "2006." not in content
    assert "2030" in content
    assert "2030." not in content
    assert "2035" in content
    assert "2035." not in content
    assert "2036" not in content
    assert "2036." not in content


def test_wrangling_magicc_input_files_error(tmpdir, caplog, test_cmip6_crunch_output):
    input_dir = test_cmip6_crunch_output
    output_dir = str(join(tmpdir, "new-sub-dir"))

    runner = CliRunner()
    with caplog.at_level("INFO"):
        result = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [
                input_dir,
                output_dir,
                "test-defaults",
                "--out-format",
                "magicc-input-files-point-end-year",
                "--regexp",
                ".*lai.*",
                "--drs",
                "CMIP6Output",
                "--number-workers",
                1,
            ],
        )
    assert result.exit_code != 0

    assert (
        str(
            "KeyError: "
            + '"'
            + "I don't know which MAGICC variable to use for input `lai`"
            + '"'
        )
        in result.output
    )


def test_wrangling_blend_models(tmpdir, caplog, test_cmip6_crunch_output):
    input_dir = test_cmip6_crunch_output
    output_dir = str(tmpdir)

    runner = CliRunner()
    with caplog.at_level("INFO"):
        result = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [
                input_dir,
                output_dir,
                "test",
                "--drs",
                "CMIP6Output",
                "--out-format",
                "tuningstrucs-blend-model",
                # avoid thetao which we can't wrangle yet
                "--regexp",
                "^((?!thetao).)*$",
            ],
        )
    assert result.exit_code == 0, result.output

    assert "netcdf-scm: {}".format(netcdf_scm.__version__) in result.output

    assert "Wrangling re.compile" in result.output
    assert ".*" in result.output
    assert ".mat" in result.output

    assert len(glob(join(output_dir, "*.mat"))) > 100


def test_wrangling_handles_integer_units(tmpdir, caplog, test_cmip6_crunch_output):
    input_dir = test_cmip6_crunch_output
    output_dir = str(tmpdir)

    runner = CliRunner()
    with caplog.at_level("INFO"):
        result = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [
                input_dir,
                output_dir,
                "test",
                "--regexp",
                ".*lai.*",
                "--out-format",
                "tuningstrucs-blend-model",
                "--drs",
                "CMIP6Output",
            ],
        )
    assert result.exit_code == 0

    assert "netcdf-scm: {}".format(netcdf_scm.__version__) in result.output

    assert "lai" in result.output


@pytest.mark.parametrize("out_format", ["mag-files", "tuningstrucs-blend-model"])
def test_wrangling_force(tmpdir, caplog, out_format, test_cmip6_crunch_output):
    input_dir = test_cmip6_crunch_output
    output_dir = str(tmpdir)

    runner = CliRunner()
    result = runner.invoke(
        wrangle_netcdf_scm_ncs,
        [
            input_dir,
            output_dir,
            "test",
            "--regexp",
            ".*lai.*",
            "-f",
            "--prefix",
            "test-prefix",
            "--drs",
            "CMIP6Output",
            "--number-workers",
            1,
            "--out-format",
            out_format,
        ],
    )
    assert result.exit_code == 0

    caplog.clear()
    with caplog.at_level("INFO"):
        result_skip = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [
                input_dir,
                output_dir,
                "test",
                "--regexp",
                ".*lai.*",
                "--prefix",
                "test-prefix",
                "--drs",
                "CMIP6Output",
                "--number-workers",
                1,
                "--out-format",
                out_format,
            ],
        )
    assert result_skip.exit_code == 0

    if out_format == "mag-files":
        expected_file = join(
            output_dir,
            "CMIP6/CMIP/CNRM-CERFACS/CNRM-CM6-1/historical/r1i1p1f2/Lmon/lai/gr/v20180917/netcdf-scm_lai_Lmon_CNRM-CM6-1_historical_r1i1p1f2_gr_200201-200512.MAG",
        )
    else:
        expected_file = join(
            output_dir,
            "test-prefix_LAI_HISTORICAL_R1I1P1F2_WORLD_SOUTHERN_HEMISPHERE_LAND.mat",
        )
    skip_str_file = "Skipped (already exists, not overwriting) {}".format(expected_file)
    assert skip_str_file in result_skip.output

    caplog.clear()
    with caplog.at_level("INFO"):
        result_force = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [
                input_dir,
                output_dir,
                "test",
                "--regexp",
                ".*lai.*",
                "-f",
                "--prefix",
                "test-prefix",
                "--drs",
                "CMIP6Output",
                "--number-workers",
                1,
                "--out-format",
                out_format,
            ],
        )
    assert result_force.exit_code == 0
    assert skip_str_file not in result_force.output


def test_wrangling_blended_models_default_drs_error(tmpdir, test_cmip6_crunch_output):
    input_dir = test_cmip6_crunch_output
    output_dir = str(tmpdir)

    runner = CliRunner()
    result = runner.invoke(
        wrangle_netcdf_scm_ncs,
        [input_dir, output_dir, "test", "--out-format", "tuningstrucs-blend-model"],
    )
    assert result.exit_code != 0
    assert isinstance(result.exception, NotImplementedError)
    assert str(result.exception) == (
        "`drs` == 'None' is not supported yet. Please raise an issue at "
        "gitlab.com/netcdf-scm/netcdf-scm/ with your use case if you need this feature."
    )


def test_wrangling_drs_replication(tmpdir, test_cmip6_crunch_output):
    input_dir = join(test_cmip6_crunch_output, "Lmon/CMIP6/CMIP/CNRM-CERFACS")
    output_dir = str(tmpdir)

    runner = CliRunner()
    result = runner.invoke(
        wrangle_netcdf_scm_ncs,
        [
            input_dir,
            output_dir,
            "test",
            "--regexp",
            ".*lai.*",
            "--drs",
            "CMIP6Output",
            "--number-workers",
            1,
        ],
    )
    assert result.exit_code == 0
    assert isdir(join(output_dir, "CMIP6/CMIP/CNRM-CERFACS"))


def test_wrangling_annual_mean_file(tmpdir, test_data_root_dir):
    input_dir = join(
        test_data_root_dir,
        "marble-cmip5-annual-output/cmip5/rcp26/Amon/tas/bcc-csm1-1/r1i1p1",
    )
    output_dir = str(tmpdir)

    runner = CliRunner()
    result = runner.invoke(
        wrangle_netcdf_scm_ncs,
        [input_dir, output_dir, "test", "--drs", "MarbleCMIP5", "--number-workers", 1],
    )

    assert result.exit_code != 0
    assert isinstance(result.exception, SystemExit)
    assert (
        "ValueError: Please raise an issue at "
        "gitlab.com/netcdf-scm/netcdf-scm/issues to discuss how to handle "
        "non-monthly data"
    ) in result.output


@pytest.mark.parametrize(
    "target_unit,conv_factor",
    (["kg / m**2 / yr", 3.1557600e07], ["g / m**2 / s", 1e03]),
)
def test_wrangling_units_specs(
    tmpdir, test_cmip6_crunch_output, target_unit, conv_factor, caplog
):
    target_units = pd.DataFrame(
        [["fgco2", target_unit], ["tos", "K"]], columns=["variable", "unit"]
    )
    target_units_csv = join(tmpdir, "target_units.csv")
    target_units.to_csv(target_units_csv, index=False)

    runner = CliRunner()

    input_dir = join(test_cmip6_crunch_output, "Omon/CMIP6/CMIP/CCCma")
    output_dir = str(tmpdir)

    caplog.clear()
    with caplog.at_level("INFO"):
        result_raw = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [
                input_dir,
                output_dir,
                "test",
                "--drs",
                "CMIP6Output",
                "--number-workers",
                1,
            ],
        )
    assert result_raw.exit_code == 0
    assert "Converting units" not in caplog.messages

    expected_file = join(
        output_dir,
        "CMIP6/CMIP/CCCma/CanESM5/piControl/r1i1p1f1/Omon/fgco2/gn/v20190429/netcdf-scm_fgco2_Omon_CanESM5_piControl_r1i1p1f1_gn_600101-600103.MAG",
    )

    res_raw = MAGICCData(expected_file)
    caplog.clear()
    with caplog.at_level("INFO"):
        result = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [
                input_dir,
                output_dir,
                "test",
                "--drs",
                "CMIP6Output",
                "--number-workers",
                1,
                "--target-units-specs",
                target_units_csv,
                "--force",
            ],
        )
    assert result.exit_code == 0
    assert (
        "Converting units of fgco2 from kg m^-2 s^-1 to {}".format(target_unit)
        in caplog.messages
    )

    res = MAGICCData(expected_file)

    np.testing.assert_allclose(
        res_raw.timeseries() * conv_factor, res.timeseries(), rtol=1e-5
    )


@pytest.mark.parametrize(
    "target_variable,original_unit,target_unit,non_area_conv_factor,input_subdir,expected_file_subpath",
    (
        (
            "fgco2",
            "kg m^-2 s^-1",
            "Gt / yr",
            10 ** -12 * 3.1557600e07,  # mass conv x time conv
            "Omon/CMIP6/CMIP/CCCma",
            "CMIP6/CMIP/CCCma/CanESM5/piControl/r1i1p1f1/Omon/fgco2/gn/v20190429/netcdf-scm_fgco2_Omon_CanESM5_piControl_r1i1p1f1_gn_600101-600103.MAG",
        ),
        (
            "hfds",
            "W m^-2",
            "W",
            1,  # only area conversion
            "Omon/CMIP6/ScenarioMIP/CCCma",
            "CMIP6/ScenarioMIP/CCCma/CanESM5/ssp119/r1i1p1f1/Omon/hfds/gn/v20190429/netcdf-scm_hfds_Omon_CanESM5_ssp119_r1i1p1f1_gn_201707-201709.MAG",
        ),
    ),
)
def test_wrangling_units_specs_area_sum(
    tmpdir,
    test_cmip6_crunch_output,
    caplog,
    target_variable,
    original_unit,
    target_unit,
    non_area_conv_factor,
    input_subdir,
    expected_file_subpath,
):
    target_units = pd.DataFrame(
        [[target_variable, target_unit], ["tos", "K"]], columns=["variable", "unit"]
    )
    target_units_csv = join(tmpdir, "target_units.csv")
    target_units.to_csv(target_units_csv, index=False)

    runner = CliRunner()

    input_dir = join(test_cmip6_crunch_output, input_subdir)
    output_dir = str(tmpdir)

    result_raw = runner.invoke(
        wrangle_netcdf_scm_ncs,
        [input_dir, output_dir, "test", "--drs", "CMIP6Output", "--number-workers", 1],
    )

    expected_file = join(output_dir, expected_file_subpath)
    assert result_raw.exit_code == 0
    res_raw = MAGICCData(expected_file)

    caplog.clear()
    with caplog.at_level("INFO"):
        result = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [
                input_dir,
                output_dir,
                "test",
                "--drs",
                "CMIP6Output",
                "--number-workers",
                1,
                "--target-units-specs",
                target_units_csv,
                "--force",
            ],
        )

    assert result.exit_code == 0
    assert (
        "Converting units of {} from {} to {}".format(
            target_variable, original_unit, target_unit
        )
        in caplog.messages
    )
    res = MAGICCData(expected_file)

    assert sorted(res["region"].tolist()) == sorted(res_raw["region"].tolist())
    for region, df in res.timeseries().groupby("region"):
        for k, v in res.metadata.items():
            if "{} (".format(ScmCube._convert_region_to_area_key(region)) in k:
                unit = k.split("(")[-1].split(")")[0]
                assert unit == "m**2", "assumed unit for test has changed..."
                conv_factor = float(v) * non_area_conv_factor
                break

        np.testing.assert_allclose(
            df.values, res_raw.filter(region=region).values * conv_factor, rtol=1e-5
        )


def test_wrangling_mag_file(tmpdir, test_cmip6_crunch_output, caplog):
    runner = CliRunner()

    input_dir = join(
        test_cmip6_crunch_output, "Lmon/CMIP6/ScenarioMIP/IPSL/IPSL-CM6A-LR"
    )
    output_dir = str(tmpdir)

    caplog.clear()
    with caplog.at_level("INFO"):
        result_raw = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [
                input_dir,
                output_dir,
                "test",
                "--drs",
                "CMIP6Output",
                "--number-workers",
                1,
            ],
        )
    assert result_raw.exit_code == 0

    expected_file = join(
        output_dir,
        "CMIP6/ScenarioMIP/IPSL/IPSL-CM6A-LR/ssp126/r1i1p1f1/Lmon/cSoilFast/gr/v20190121/netcdf-scm_cSoilFast_Lmon_IPSL-CM6A-LR_ssp126_r1i1p1f1_gr_202501-204012.MAG",
    )

    with open(expected_file) as f:
        content = f.read()

    assert "THISFILE_TIMESERIESTYPE = 'MONTHLY'" in content


def _get_expected_wrangled_ts(res_raw, out_format_mag):
    if out_format_mag.endswith("average-year-start-year"):
        # drop out last year as we don't want the wrangler to add an extra year to the
        # data
        return res_raw.time_mean("AS").timeseries().iloc[:, :-1]

    if out_format_mag.endswith("average-year-mid-year"):
        # drop out last year as we don't want the wrangler to add an extra year to the
        # data
        return res_raw.time_mean("AC").timeseries()

    if out_format_mag.endswith("average-year-end-year"):
        # drop out first year as we don't want the wrangler to add an extra year to the
        # data
        return res_raw.time_mean("A").timeseries().iloc[:, 1:]

    if out_format_mag.endswith("point-start-year"):
        # drop out last year as we don't want the wrangler to add an extra year to the
        # data
        return res_raw.resample("AS").timeseries().iloc[:, :-1]

    if out_format_mag.endswith("point-mid-year"):
        # drop out last year as we don't want the wrangler to add an extra year to the
        # data
        out_time_points = [
            dt.datetime(y, 7, 1)
            for y in range(res_raw["time"].min().year, res_raw["time"].max().year + 1,)
        ]
        return res_raw.interpolate(target_times=out_time_points).timeseries()

    if out_format_mag.endswith("point-end-year"):
        # drop out first year as we don't want the wrangler to add an extra year to the
        # data
        return res_raw.resample("A").timeseries().iloc[:, 1:]

    raise AssertionError("shouldn't get here")


@pytest.mark.parametrize(
    "out_format_mag",
    (
        "mag-files-average-year-start-year",
        "mag-files-average-year-mid-year",
        "mag-files-average-year-end-year",
        "mag-files-point-start-year",
        "mag-files-point-mid-year",
        "mag-files-point-end-year",
    ),
)
def test_wrangling_mag_file_operations(
    tmpdir, test_cmip6_crunch_output, caplog, out_format_mag
):
    runner = CliRunner()

    input_dir = join(
        test_cmip6_crunch_output, "Lmon/CMIP6/ScenarioMIP/IPSL/IPSL-CM6A-LR"
    )
    output_dir = str(tmpdir)

    caplog.clear()
    with caplog.at_level("INFO"):
        result_raw = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [
                input_dir,
                output_dir,
                "test",
                "--drs",
                "CMIP6Output",
                "--number-workers",
                1,
            ],
        )
    assert result_raw.exit_code == 0

    expected_file_raw = join(
        output_dir,
        "CMIP6/ScenarioMIP/IPSL/IPSL-CM6A-LR/ssp126/r1i1p1f1/Lmon/cSoilFast/gr/v20190121/netcdf-scm_cSoilFast_Lmon_IPSL-CM6A-LR_ssp126_r1i1p1f1_gr_202501-204012.MAG",
    )

    res_raw = MAGICCData(expected_file_raw)
    res_raw_resampled = _get_expected_wrangled_ts(res_raw, out_format_mag)

    caplog.clear()
    with caplog.at_level("INFO"):
        result = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [
                input_dir,
                output_dir,
                "test",
                "--drs",
                "CMIP6Output",
                "--out-format",
                out_format_mag,
                "--number-workers",
                1,
            ],
        )
    assert result.exit_code == 0

    expected_file = join(
        output_dir,
        "CMIP6/ScenarioMIP/IPSL/IPSL-CM6A-LR/ssp126/r1i1p1f1/Lmon/cSoilFast/gr/v20190121/netcdf-scm_cSoilFast_Lmon_IPSL-CM6A-LR_ssp126_r1i1p1f1_gr_2025-2040.MAG",
    )

    res = MAGICCData(expected_file)

    np.testing.assert_allclose(res_raw_resampled, res.timeseries(), rtol=2 * 1e-3)
    with open(expected_file) as f:
        content = f.read()

    assert (
        "THISFILE_TIMESERIESTYPE = '{}'".format(
            out_format_mag.replace("mag-files-", "").replace("-", "_").upper()
        )
        in content
    )


def test_wrangling_in_file(tmpdir, test_cmip6_crunch_output, caplog):
    runner = CliRunner()

    input_dir = join(
        test_cmip6_crunch_output, "Amon/CMIP6/CMIP/IPSL/IPSL-CM6A-LR/historical"
    )
    output_dir = str(tmpdir)

    caplog.clear()
    with caplog.at_level("INFO"):
        result_raw = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [
                input_dir,
                output_dir,
                "test",
                "--drs",
                "CMIP6Output",
                "--number-workers",
                1,
                "--out-format",
                "magicc-input-files",
                "--regexp",
                ".*tas.*",
            ],
        )
    assert result_raw.exit_code == 0, result_raw.stdout

    # also a global file but don't worry about that
    expected_file = join(
        output_dir,
        "CMIP6/CMIP/IPSL/IPSL-CM6A-LR/historical/r1i1p1f1/Amon/tas/gr/v20180803/TAS_HISTORICAL_IPSL-CM6A-LR_R1I1P1F1_FOURBOX_SURFACE_TEMP.IN",
    )

    with open(expected_file) as f:
        content = f.read()

    assert "timeseriestype: MONTHLY" in content


@pytest.mark.parametrize(
    "out_format_in_file",
    (
        "magicc-input-files-average-year-start-year",
        "magicc-input-files-average-year-mid-year",
        "magicc-input-files-average-year-end-year",
        "magicc-input-files-point-start-year",
        "magicc-input-files-point-mid-year",
        "magicc-input-files-point-end-year",
    ),
)
def test_wrangling_in_file_operations(
    tmpdir, test_cmip6_crunch_output, caplog, out_format_in_file
):
    runner = CliRunner()

    input_dir = join(
        test_cmip6_crunch_output, "Amon/CMIP6/CMIP/IPSL/IPSL-CM6A-LR/piControl"
    )
    output_dir = str(tmpdir)

    caplog.clear()
    with caplog.at_level("INFO"):
        result_raw = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [
                input_dir,
                output_dir,
                "test",
                "--drs",
                "CMIP6Output",
                "--number-workers",
                1,
                "--out-format",
                "magicc-input-files",
                "--regexp",
                ".*tas.*",
            ],
        )
    assert result_raw.exit_code == 0, result_raw.stdout

    # also a global file but don't worry about that
    expected_file_raw = join(
        output_dir,
        "CMIP6/CMIP/IPSL/IPSL-CM6A-LR/piControl/r1i1p1f1/Amon/tas/gr/v20181123/TAS_PICONTROL_IPSL-CM6A-LR_R1I1P1F1_FOURBOX_SURFACE_TEMP.IN",
    )

    res_raw = MAGICCData(expected_file_raw)
    res_raw_resampled = _get_expected_wrangled_ts(res_raw, out_format_in_file)

    caplog.clear()
    with caplog.at_level("INFO"):
        result = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [
                input_dir,
                output_dir,
                "test",
                "--drs",
                "CMIP6Output",
                "--out-format",
                out_format_in_file,
                "--regexp",
                ".*tas.*",
                "--number-workers",
                1,
                "-f",
            ],
        )
    assert result.exit_code == 0, result.stdout

    os.listdir(
        join(
            output_dir,
            "CMIP6/CMIP/IPSL/IPSL-CM6A-LR/piControl/r1i1p1f1/Amon/tas/gr/v20181123/",
        )
    )
    expected_file = join(
        output_dir,
        "CMIP6/CMIP/IPSL/IPSL-CM6A-LR/piControl/r1i1p1f1/Amon/tas/gr/v20181123/TAS_PICONTROL_IPSL-CM6A-LR_R1I1P1F1_FOURBOX_SURFACE_TEMP.IN",
    )

    res = MAGICCData(expected_file)

    np.testing.assert_allclose(res_raw_resampled, res.timeseries(), rtol=2 * 1e-3)
    with open(expected_file) as f:
        content = f.read()

    assert (
        "timeseriestype: {}".format(
            out_format_in_file.replace("magicc-input-files-", "")
            .replace("-", "_")
            .upper()
        )
        in content
    )


@pytest.mark.xfail(
    message="Some of these errors are still cryptic (and they shouldn't be) because scmdata isn't interpolating as intended"
)
@pytest.mark.parametrize(
    "out_format",
    (
        "mag-files-average-year-start-year",
        "mag-files-average-year-mid-year",
        "mag-files-average-year-end-year",
        "mag-files-point-start-year",
        "mag-files-point-mid-year",
        "mag-files-point-end-year",
        "magicc-input-files-average-year-start-year",
        "magicc-input-files-average-year-mid-year",
        "magicc-input-files-average-year-end-year",
        "magicc-input-files-point-start-year",
        "magicc-input-files-point-mid-year",
        "magicc-input-files-point-end-year",
    ),
)
def test_wrangling_with_operation_single_timepoint(
    tmpdir, test_cmip6_crunch_output, caplog, out_format
):
    runner = CliRunner()

    input_dir = join(
        test_cmip6_crunch_output,
        "Amon/CMIP6/CMIP/NCAR/CESM2/historical/r7i1p1f1/Amon/tas",
    )
    output_dir = str(tmpdir)

    caplog.clear()
    with caplog.at_level("INFO"):
        result = runner.invoke(
            wrangle_netcdf_scm_ncs,
            [
                input_dir,
                output_dir,
                "test",
                "--drs",
                "CMIP6Output",
                "--number-workers",
                1,
                "--out-format",
                out_format,
            ],
        )
    assert result.exit_code != 0
    assert (
        "ValueError: We cannot yet write `{}` if the output data will have only one timestep".format(
            out_format
        )
        in result.stdout
    )



import glob
import shutil
import warnings
from os import makedirs, path, walk
from os.path import abspath, dirname, isdir, join

import cf_units as unit
import iris
import numpy as np
import pandas as pd
import pytest
from pymagicc.io import MAGICCData
from scmdata import ScmDataFrame

from netcdf_scm.io import load_scmdataframe
from netcdf_scm.iris_cube_wrappers import (
    CMIP6Input4MIPsCube,
    CMIP6OutputCube,
    MarbleCMIP5Cube,
    ScmCube,
)

TEST_DATA_ROOT_DIR = join(dirname(abspath(__file__)), "test-data")


TEST_STITCH_OUTPUT_DIR = join(TEST_DATA_ROOT_DIR, "expected-stitch-output")


@pytest.fixture(scope="session")
def test_data_root_dir():
    if not isdir(TEST_DATA_ROOT_DIR):
        pytest.skip("test data required")
    return TEST_DATA_ROOT_DIR


@pytest.fixture(scope="session")
def test_data_knmi_dir(test_data_root_dir):
    return join(test_data_root_dir, "knmi-climate-explorer")


@pytest.fixture(scope="session")
def test_data_marble_cmip5_dir(test_data_root_dir):
    return join(test_data_root_dir, "marble-cmip5")


@pytest.fixture(scope="session")
def test_tas_file(test_data_marble_cmip5_dir):
    return join(
        test_data_marble_cmip5_dir,
        "cmip5",
        "1pctCO2",
        "Amon",
        "tas",
        "CanESM2",
        "r1i1p1",
        "tas_Amon_CanESM2_1pctCO2_r1i1p1_189201-190312.nc",
    )


@pytest.fixture(scope="session")
def test_sftlf_file(test_data_marble_cmip5_dir):
    return join(
        test_data_marble_cmip5_dir,
        "cmip5",
        "1pctCO2",
        "fx",
        "sftlf",
        "CanESM2",
        "r0i0p0",
        "sftlf_fx_CanESM2_1pctCO2_r0i0p0.nc",
    )


@pytest.fixture(scope="session")
def test_areacella_file(test_data_marble_cmip5_dir):
    return join(
        test_data_marble_cmip5_dir,
        "cmip5",
        "1pctCO2",
        "fx",
        "areacella",
        "CanESM2",
        "r0i0p0",
        "areacella_fx_CanESM2_1pctCO2_r0i0p0.nc",
    )


@pytest.fixture(scope="session")
def test_access_cmip5_file(test_data_marble_cmip5_dir):
    return join(
        test_data_marble_cmip5_dir,
        "cmip5",
        "rcp45",
        "Amon",
        "tas",
        "ACCESS1-0",
        "r1i1p1",
        "tas_Amon_ACCESS1-0_rcp45_r1i1p1_200601-201012.nc",
    )


@pytest.fixture(scope="session")
def test_marble_cmip5_output_tas_file(test_data_marble_cmip5_dir):
    return join(
        test_data_marble_cmip5_dir,
        "cmip5",
        "historical",
        "Amon",
        "tas",
        "ACCESS1-0",
        "r1i1p1",
        "tas_Amon_ACCESS1-0_historical_r1i1p1_187701-187703.nc",
    )


@pytest.fixture(scope="session")
def test_marble_cmip5_output_hfds_file(test_data_marble_cmip5_dir):
    return join(
        test_data_marble_cmip5_dir,
        "cmip5",
        "historical",
        "Omon",
        "hfds",
        "ACCESS1-0",
        "r1i1p1",
        "hfds_Omon_ACCESS1-0_historical_r1i1p1_187701-187703.nc",
    )


@pytest.fixture(scope="session")
def test_data_cmip6input4mips_dir(test_data_root_dir):
    return join(test_data_root_dir, "cmip6input4mips")


@pytest.fixture(scope="session")
def test_cmip6input4mips_historical_concs_file(test_data_cmip6input4mips_dir):
    return join(
        test_data_cmip6input4mips_dir,
        "input4MIPs",
        "CMIP6",
        "CMIP",
        "UoM",
        "UoM-CMIP-1-2-0",
        "atmos",
        "yr",
        "mole-fraction-of-so2f2-in-air",
        "gr1-GMNHSH",
        "v20100304",
        "mole-fraction-of-so2f2-in-air_input4MIPs_GHGConcentrations_CMIP_UoM-CMIP-1-2-0_gr1-GMNHSH_0000-2014.nc",
    )


@pytest.fixture(scope="session")
def test_cmip6input4mips_projection_concs_file(test_data_cmip6input4mips_dir):
    return join(
        test_data_cmip6input4mips_dir,
        "input4MIPs",
        "CMIP6",
        "ScenarioMIP",
        "UoM",
        "UoM-MESSAGE-GLOBIOM-ssp245-1-2-1",
        "atmos",
        "mon",
        "mole-fraction-of-carbon-dioxide-in-air",
        "gn-15x360deg",
        "v20100304",
        "mole-fraction-of-carbon-dioxide-in-air_input4MIPs_GHGConcentrations_ScenarioMIP_UoM-MESSAGE-GLOBIOM-ssp245-1-2-1_gn-15x360deg_201501-203012.nc",
    )


TEST_DATA_CMIP6OUTPUT_DIR = join(TEST_DATA_ROOT_DIR, "cmip6output")


@pytest.fixture(scope="session")
def test_data_cmip6output_dir(test_data_root_dir):
    return TEST_DATA_CMIP6OUTPUT_DIR


@pytest.fixture(scope="session")
def test_cmip6_output_file(test_data_cmip6output_dir):
    return join(
        test_data_cmip6output_dir,
        "CMIP6",
        "CMIP",
        "BCC",
        "BCC-CSM2-MR",
        "1pctCO2",
        "r1i1p1f1",
        "Amon",
        "rlut",
        "gn",
        "v20181015",
        "rlut_Amon_BCC-CSM2-MR_1pctCO2_r1i1p1f1_gn_185001-185912.nc",
    )


@pytest.fixture(scope="session")
def test_cmip6_output_file_missing_bounds(test_data_cmip6output_dir):
    return join(
        test_data_cmip6output_dir,
        "CMIP6",
        "ScenarioMIP",
        "IPSL",
        "IPSL-CM6A-LR",
        "ssp126",
        "r1i1p1f1",
        "Lmon",
        "cSoilFast",
        "gr",
        "v20190121",
        "cSoilFast_Lmon_IPSL-CM6A-LR_ssp126_r1i1p1f1_gr_202501-204012.nc",
    )


@pytest.fixture(scope="session")
def test_cmip6_output_file_1_unit(test_data_cmip6output_dir):
    return join(
        test_data_cmip6output_dir,
        "CMIP6",
        "CMIP",
        "CNRM-CERFACS",
        "CNRM-CM6-1",
        "historical",
        "r1i1p1f2",
        "Lmon",
        "lai",
        "gr",
        "v20180917",
        "lai_Lmon_CNRM-CM6-1_historical_r1i1p1f2_gr_200201-200512.nc",
    )


@pytest.fixture(scope="session")
def test_cmip6_output_tas_file(test_data_cmip6output_dir):
    return join(
        test_data_cmip6output_dir,
        "CMIP6",
        "CMIP",
        "NCAR",
        "CESM2",
        "historical",
        "r7i1p1f1",
        "Amon",
        "tas",
        "gn",
        "v20190311",
        "tas_Amon_CESM2_historical_r7i1p1f1_gn_195701-195703.nc",
    )


@pytest.fixture(scope="session")
def test_cmip6_output_fgco2_file(test_data_cmip6output_dir):
    return join(
        test_data_cmip6output_dir,
        "CMIP6",
        "CMIP",
        "CCCma",
        "CanESM5",
        "piControl",
        "r1i1p1f1",
        "Omon",
        "fgco2",
        "gn",
        "v20190429",
        "fgco2_Omon_CanESM5_piControl_r1i1p1f1_gn_600101-600103.nc",
    )


TEST_CMIP6OUTPUT_HFDS_FILE = join(
    TEST_DATA_CMIP6OUTPUT_DIR,
    "CMIP6",
    "CMIP",
    "NCAR",
    "CESM2",
    "historical",
    "r7i1p1f1",
    "Omon",
    "hfds",
    "gr",
    "v20190311",
    "hfds_Omon_CESM2_historical_r7i1p1f1_gr_195701-195703.nc",
)


@pytest.fixture(scope="session")
def test_cmip6_output_hfds_file():
    return TEST_CMIP6OUTPUT_HFDS_FILE


TEST_CMIP6OUTPUT_HFDS_NATIVE_GRID_FILE = TEST_CMIP6OUTPUT_HFDS_FILE.replace("gr", "gn")


@pytest.fixture(scope="session")
def test_cmip6_output_hfds_native_grid_file():
    return TEST_CMIP6OUTPUT_HFDS_NATIVE_GRID_FILE


TEST_CMIP6OUTPUT_HFDS_WITH_NLON_NLAT_COORDS_FILE = join(
    TEST_DATA_CMIP6OUTPUT_DIR,
    "CMIP6",
    "GeoMIP",
    "NCAR",
    "CESM2-WACCM",
    "G6solar",
    "r1i1p1f1",
    "Omon",
    "hfds",
    "gn",
    "v20191001",
    "hfds_Omon_CESM2-WACCM_G6solar_r1i1p1f1_gn_202107-202109.nc",
)


@pytest.fixture(scope="session")
def test_cmip6_output_hfds_nlon_nlat_coords_file():
    return TEST_CMIP6OUTPUT_HFDS_WITH_NLON_NLAT_COORDS_FILE


TEST_CMIP6OUTPUT_HFDS_WITH_I_J_COORDS_FILE = join(
    TEST_DATA_CMIP6OUTPUT_DIR,
    "CMIP6",
    "ScenarioMIP",
    "CCCma",
    "CanESM5",
    "ssp119",
    "r1i1p1f1",
    "Omon",
    "hfds",
    "gn",
    "v20190429",
    "hfds_Omon_CanESM5_ssp119_r1i1p1f1_gn_201707-201709.nc",
)


@pytest.fixture(scope="session")
def test_cmip6_output_hfds_i_j_coords_file():
    return TEST_CMIP6OUTPUT_HFDS_WITH_I_J_COORDS_FILE


TEST_CMIP6OUTPUT_HFDS_WITH_VERTEX_COORDS_FILE = join(
    TEST_DATA_CMIP6OUTPUT_DIR,
    "CMIP6",
    "DAMIP",
    "CCCma",
    "CanESM5",
    "hist-stratO3",
    "r1i1p1f1",
    "Omon",
    "fgco2",
    "gn",
    "v20190429",
    "fgco2_Omon_CanESM5_hist-stratO3_r1i1p1f1_gn_185207-185209.nc",
)


@pytest.fixture(scope="session")
def test_cmip6_output_hfds_vertex_coords_file():
    return TEST_CMIP6OUTPUT_HFDS_WITH_VERTEX_COORDS_FILE


TEST_CMIP6OUTPUT_SFTLF_GN_FILE = join(
    TEST_DATA_CMIP6OUTPUT_DIR,
    "CMIP6",
    "CMIP",
    "NASA-GISS",
    "GISS-E2-1-G",
    "abrupt-4xCO2",
    "r1i1p1f1",
    "Amon",
    "tas",
    "gn",
    "v20181002",
    "tas_Amon_GISS-E2-1-G_abrupt-4xCO2_r1i1p1f1_gn_185001-185412.nc",
)


@pytest.fixture(scope="session")
def test_cmip6_output_sftlf_gn_file():
    return TEST_CMIP6OUTPUT_SFTLF_GN_FILE


@pytest.fixture(
    scope="session",
    params=[TEST_CMIP6OUTPUT_HFDS_FILE, TEST_CMIP6OUTPUT_HFDS_NATIVE_GRID_FILE],
)
def test_cmip6_output_hfds_files(request):
    return request.param


@pytest.fixture(scope="session")
def test_cmip6_output_thetao_file(test_data_cmip6output_dir):
    return join(
        test_data_cmip6output_dir,
        "CMIP6",
        "CMIP",
        "NCAR",
        "CESM2",
        "historical",
        "r10i1p1f1",
        "Omon",
        "thetao",
        "gn",
        "v20190313",
        "thetao_Omon_CESM2_historical_r10i1p1f1_gn_195310-195312.nc",
    )


@pytest.fixture(scope="session")
def test_cmip6_output_hfds_concatenate_directory(test_data_cmip6output_dir):
    return join(
        test_data_cmip6output_dir,
        "CMIP6",
        "CMIP",
        "NCAR",
        "CESM2",
        "historical",
        "r10i1p1f1",
        "Omon",
        "tos",
        "gn",
        "v20190313",
    )


@pytest.fixture(scope="session")
def test_cmip6_output_conflicting_phenomenon_metadata_directory(
    test_data_cmip6output_dir,
):
    return join(
        test_data_cmip6output_dir,
        "CMIP6",
        "CMIP",
        "NCAR",
        "CESM2",
        "abrupt-4xCO2",
        "r1i1p1f1",
        "Omon",
        "fgco2",
        "gr",
        "v20190927",
    )


@pytest.fixture(scope="session")
def test_cmip6_output_ocean_data_handling_directory(test_data_cmip6output_dir):
    return join(
        test_data_cmip6output_dir,
        "CMIP6",
        "ScenarioMIP",
        "CCCma",
        "CanESM5",
        "ssp119",
        "r1i1p1f1",
        "Omon",
        "hfds",
        "gn",
        "v20190429",
    )


@pytest.fixture(scope="session")
def test_cmip6_output_different_grid_same_answer_issue_35(test_data_cmip6output_dir):
    gr = join(
        test_data_cmip6output_dir,
        "CMIP6",
        "CMIP",
        "NCAR",
        "CESM2",
        "1pctCO2",
        "r1i1p1f1",
        "Omon",
        "fgco2",
        "gr",
        "v20190425",
    )
    gn = gr.replace("gr", "gn")

    return {"gr": gr, "gn": gn}


@pytest.fixture(scope="session")
def test_cmip6_crunch_output(test_data_root_dir):
    return join(test_data_root_dir, "expected-crunching-output", "cmip6output")


@pytest.fixture(scope="session")
def test_cmip6_wrangle_output(test_data_root_dir):
    return join(test_data_root_dir, "expected-wrangling-output", "cmip6output")


@pytest.fixture(scope="session")
def test_cmip6_stitch_output(test_data_root_dir):
    return join(TEST_STITCH_OUTPUT_DIR, "cmip6output")


@pytest.fixture(
    params=[
        f
        for f in glob.glob(join(TEST_STITCH_OUTPUT_DIR, "**", "*.MAG"), recursive=True)
    ]
)
def stitched_mag_file(request):
    yield request.param


@pytest.fixture(scope="session")
def test_marble_cmip5_crunch_output(test_data_root_dir):
    return join(
        test_data_root_dir, "expected-crunching-output", "marble-cmip5", "cmip5"
    )


@pytest.fixture(scope="session")
def test_marble_cmip5_stitch_output(test_data_root_dir):
    return join(TEST_STITCH_OUTPUT_DIR, "marble-cmip5output")


@pytest.fixture(scope="session")
def test_data_netcdfscm_ncs_dir(test_data_root_dir):
    return join(test_data_root_dir, "netcdf-scm-ncs")


@pytest.fixture(scope="session")
def test_data_netcdfscm_nc_file(test_data_root_dir):
    return join(
        test_data_root_dir,
        "netcdf-scm_tas_Amon_ACCESS1-0_rcp45_r1i1p1_200601-204912.nc",
    )


TEST_DATA_HISTORICAL_CONC_GMNHSH_FILE = join(
    TEST_DATA_ROOT_DIR,
    "mole-fraction-of-carbon-dioxide-in-air_input4MIPs_GHGConcentrations_CMIP_UoM-CMIP-1-2-0_gr1-GMNHSH_000001-201412.nc",
)

TEST_DATA_PROJECTION_CONC_CO2_GMNHSH_FILE = join(
    TEST_DATA_ROOT_DIR,
    "mole-fraction-of-carbon-dioxide-in-air_input4MIPs_GHGConcentrations_ScenarioMIP_UoM-MESSAGE-GLOBIOM-ssp245-1-2-1_gr1-GMNHSH_201501-250012.nc",
)

TEST_DATA_PROJECTION_CONC_C4F10_GMNHSH_FILE = join(
    TEST_DATA_ROOT_DIR,
    "mole-fraction-of-c4f10-in-air_input4MIPs_GHGConcentrations_AerChemMIP_UoM-AIM-ssp370-lowNTCF-1-2-1_gr1-GMNHSH_201501-250012.nc",
)


@pytest.fixture(
    scope="session",
    params=[
        TEST_DATA_HISTORICAL_CONC_GMNHSH_FILE,
        TEST_DATA_PROJECTION_CONC_CO2_GMNHSH_FILE,
        TEST_DATA_PROJECTION_CONC_C4F10_GMNHSH_FILE,
    ],
)
def test_data_conc_gmnhsh_file(request):
    return request.param


def pytest_addoption(parser):
    parser.addoption(
        "--update-expected-files",
        action="store_true",
        default=False,
        help="Overwrite expected files",
    )


@pytest.fixture
def update_expected_files(request):
    return request.config.getoption("--update-expected-files")


def get_test_cube_lon():
    lon = iris.coords.DimCoord(
        np.array([45, 135, 225, 315]),
        standard_name="longitude",
        units=unit.Unit("degrees"),
        long_name="longitude",
        var_name="lon",
        circular=True,
    )
    lon.guess_bounds()
    return lon


def get_test_cube_lat():
    lat = iris.coords.DimCoord(
        np.array([60, 0, -60]),
        standard_name="latitude",
        units=unit.Unit("degrees"),
        long_name="latitude",
        var_name="lat",
    )
    lat.guess_bounds()
    return lat


def get_test_cube_time():
    return iris.coords.DimCoord(
        np.array([365, 365 * 2, 365 * 3, 365 * 3 + 180]),
        standard_name="time",
        units=unit.Unit("days since 1850-1-1", calendar="365_day"),
        long_name="time",
        var_name="time",
    )


def get_test_cube_attributes():
    return {
        "Creator": "Blinky Bill",
        "Supervisor": "Patch",
        "attribute 3": "attribute 3",
        "attribute d": "hello, attribute d",
    }


def create_cube(cube_cls):
    c = cube_cls()
    test_data = np.ma.masked_array(
        [
            [[0, 0.5, 1, 3], [0.0, 0.15, 0.25, 0.3], [-4, -5, -6, -7]],
            [[9, 9, 7, 6], [0, 1, 2, 3], [5, 4, 3, 2]],
            [[10, 14, 12, 13], [0, 1, 2, 3], [4.1, 5.2, 6.2, 7.3]],
            [[10, 18, 12, 13], [0, 1, 2, 3], [4.1, 5.2, 6.2, 7.3]],
        ],
        mask=False,
    )

    c.cube = iris.cube.Cube(
        test_data,
        standard_name="air_temperature",
        long_name="air_temperature",
        var_name="air_temperature",
        dim_coords_and_dims=[
            (get_test_cube_time(), 0),
            (get_test_cube_lat(), 1),
            (get_test_cube_lon(), 2),
        ],
        units=unit.Unit("degC"),
        attributes=get_test_cube_attributes(),
    )

    return c


@pytest.fixture(scope="function")
def test_cube(request):
    return create_cube(request.cls.tclass)


@pytest.fixture(
    scope="function",
    params=[ScmCube, MarbleCMIP5Cube, CMIP6Input4MIPsCube, CMIP6OutputCube],
)
def test_all_cubes(request):
    return create_cube(request.param)


def create_sftlf_cube(cube_cls):
    c = cube_cls()
    c._loaded_paths.append("test_sftlf.nc")

    test_data = np.ma.masked_array(
        [[90, 49.9, 50.0, 50.1], [100, 49, 50, 51], [51, 30, 10, 0]], mask=False
    )
    c.cube = iris.cube.Cube(test_data)
    c.cube.standard_name = "land_area_fraction"

    c.cube.add_dim_coord(get_test_cube_lat(), 0)
    c.cube.add_dim_coord(get_test_cube_lon(), 1)

    return c


@pytest.fixture(scope="function")
def test_sftlf_cube(request):
    return create_sftlf_cube(request.cls.tclass)


@pytest.fixture(scope="function")
def test_generic_tas_cube(test_tas_file):
    test_generic_tas_cube = ScmCube()
    # can safely ignore these warnings here
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", "Missing CF-netCDF measure variable 'areacella'"
        )
        test_generic_tas_cube.cube = iris.load_cube(test_tas_file)

    return test_generic_tas_cube


@pytest.fixture
def run_crunching_comparison(assert_scmdata_frames_allclose):
    def _do_comparison(res, expected, update=False):
        """Run test that crunched files are unchanged

        Parameters
        ----------
        res : str
            Directory written as part of the test
        expected : str
            Directory against which the comparison should be done
        update : bool
            If True, don't perform the test and instead simply
            overwrite the ``expected`` with ``res``

        Raises
        ------
        AssertionError
            If ``update`` is ``False`` and ``res`` and ``expected``
            are not identical.
        """
        paths_to_walk = [expected, res] if not update else [res]
        for p in paths_to_walk:
            for dirpath, _, filenames in walk(p):
                if filenames:
                    if update:
                        path_to_check = dirpath.replace(res, expected)
                        if not path.exists(path_to_check):
                            makedirs(path_to_check)

                    for f in filenames:
                        base_f = join(dirpath, f)
                        comparison_p = expected if p == res else res
                        comparison_f = base_f.replace(p, comparison_p)
                        assert base_f != comparison_f
                        if update:
                            print("Updating {}".format(comparison_f))
                            shutil.copy(base_f, comparison_f)
                        else:
                            try:
                                base_scmdf = load_scmdataframe(base_f)
                                comparison_scmdf = load_scmdataframe(comparison_f)
                                assert_scmdata_frames_allclose(
                                    base_scmdf, comparison_scmdf
                                )
                            except NotImplementedError:  # 3D data
                                base_cubes = iris.load(base_f)
                                comparison_cubes = iris.load(comparison_f)
                                for comparison_cube in comparison_cubes:
                                    region = comparison_cube.attributes["region"]
                                    for base_cube in base_cubes:
                                        if base_cube.attributes["region"] == region:
                                            break

                                    np.testing.assert_allclose(
                                        base_cube.data, comparison_cube.data, rtol=1e-6
                                    )
                                    base_cube.attributes.pop(
                                        "crunch_netcdf_scm_version"
                                    )
                                    comparison_cube.attributes.pop(
                                        "crunch_netcdf_scm_version"
                                    )
                                    assert (
                                        base_cube.attributes
                                        == comparison_cube.attributes
                                    )

        if update:
            pytest.skip("Updated {}".format(expected))

    return _do_comparison


@pytest.fixture
def run_wrangling_comparison(assert_scmdata_frames_allclose):
    def _do_comparison(res, expected, update=False):
        """Run test that crunched files are unchanged

        Parameters
        ----------
        res : str
            Directory written as part of the test
        expected : str
            Directory against which the comparison should be done
        update : bool
            If True, don't perform the test and instead simply
            overwrite the ``expected`` with ``res``

        Raises
        ------
        AssertionError
            If ``update`` is ``False`` and ``res`` and ``expected``
            are not identical.
        """
        paths_to_walk = [expected, res] if not update else [res]
        for p in paths_to_walk:
            for dirpath, _, filenames in walk(p):
                if filenames:
                    if update:
                        path_to_check = dirpath.replace(res, expected)
                        if not path.exists(path_to_check):
                            makedirs(path_to_check)

                    for f in filenames:
                        base_f = join(dirpath, f)
                        comparison_p = expected if p == res else res
                        comparison_f = base_f.replace(p, comparison_p)
                        assert base_f != comparison_f
                        if update:
                            print("Updating {}".format(comparison_f))
                            shutil.copy(base_f, comparison_f)
                        else:
                            base_scmdf = MAGICCData(base_f)
                            comparison_scmdf = MAGICCData(comparison_f)
                            assert_scmdata_frames_allclose(base_scmdf, comparison_scmdf)

        if update:
            pytest.skip("Updated {}".format(expected))

    return _do_comparison


@pytest.fixture
def assert_scmdata_frames_allclose():
    def _do_assertion(res_scmdf, exp_scmdf):
        res_df = res_scmdf.timeseries().sort_index()
        assert not res_df.isnull().any().any(), "Failed sanity check"
        assert (
            (res_df.values > -(10 ** 5)) & (res_df.values < 10 ** 5)
        ).all(), "Failed sanity check"

        exp_df = exp_scmdf.timeseries().sort_index()
        pd.testing.assert_frame_equal(res_df, exp_df, check_like=True)

        for base, check in [(exp_scmdf, res_scmdf), (res_scmdf, exp_scmdf)]:
            for k, v in base.metadata.items():
                variable_keys = ("date", "crunch_netcdf_scm_version", "header")
                if any([vk in k for vk in variable_keys]):
                    continue  # will change with version and test run time

                if k == "crunch_source_files":
                    assert sorted([w.strip() for w in v.split(";")]) == sorted(
                        [w.strip() for w in check.metadata[k].split(";")]
                    )
                elif isinstance(v, (np.ndarray, np.float, np.int)):
                    np.testing.assert_allclose(v, check.metadata[k], rtol=1e-6)
                else:
                    assert v == check.metadata[k]

    return _do_assertion


NORMALISATION_METHODS = [
    "31-yr-mean-after-branch-time",
    "21-yr-running-mean",
    "21-yr-running-mean-dedrift",
]


@pytest.fixture(params=NORMALISATION_METHODS)
def normalisation_method(request):
    return request.param


@pytest.fixture(scope="function")
def junk_data_normalisation():
    out = ScmDataFrame(
        np.arange(100),
        index=np.arange(1850, 1950),
        columns={
            "scenario": ["rcp26"],
            "experiment_id": ["rcp26"],
            "model": ["model1"],
            "climate_model": ["climate_model1"],
            "mip_era": ["CMIP6"],
            "region": ["World"],
            "variable": ["tas"],
            "unit": ["K"],
        },
    )

    out.metadata = {
        "branch_time_in_parent": 0.0,
        "parent_time_units": "days since 1850-01-01",
        "calendar": "365_day",
        "source_id": "UoM mocked data",
        "experiment_id": "rcp26",
        "netcdf-scm crunched file": "/path/to/mocked/rcp26/data_rcp26.nc",
    }

    return out


@pytest.fixture(scope="function")
def junk_data_picontrol_normalisation(junk_data_normalisation):
    out = junk_data_normalisation.copy()
    out.set_meta("piControl", "scenario")
    out.set_meta("piControl", "experiment_id")
    out.metadata = {
        "other": "metadata",
        "goes": "here",
        "branch_time_in_parent": -1001.1,
        "parent_time_units": "days since 1850-01-01",
        "calendar": "365_day",
        "source_id": "UoM mocked data",
        "experiment_id": "piControl",
        "netcdf-scm crunched file": "/path/to/mocked/piControl/data_piControl.nc",
    }

    return out


@pytest.fixture(scope="session")
def picontrol_data_normalisation(test_cmip6_crunch_output):
    test_file = join(
        test_cmip6_crunch_output,
        "Amon",
        "CMIP6",
        "CMIP",
        "NCAR",
        "CESM2",
        "piControl",
        "r1i1p1f1",
        "Amon",
        "tas",
        "gn",
        "v20190320",
        "netcdf-scm_tas_Amon_CESM2_piControl_r1i1p1f1_gn_080001-099912.nc",
    )

    loaded = load_scmdataframe(test_file)
    loaded.metadata["netcdf-scm crunched file"] = test_file

    return loaded


@pytest.fixture(scope="session")
def historical_data_normalisation(test_cmip6_crunch_output):
    test_file = join(
        test_cmip6_crunch_output,
        "Amon",
        "CMIP6",
        "CMIP",
        "NCAR",
        "CESM2",
        "historical",
        "r10i1p1f1",
        "Amon",
        "tas",
        "gn",
        "v20190313",
        "netcdf-scm_tas_Amon_CESM2_historical_r10i1p1f1_gn_185001-201412.nc",
    )

    loaded = load_scmdataframe(test_file)
    loaded.metadata["netcdf-scm crunched file"] = test_file

    return loaded



from os.path import join

import pytest
from click.testing import CliRunner

from netcdf_scm.cli import wrangle_netcdf_scm_ncs


@pytest.mark.parametrize(
    "out_format", (["mag-files", "mag-files-average-year-mid-year"])
)
def test_wrangling_results(
    tmpdir,
    update_expected_files,
    test_cmip6_crunch_output,
    test_cmip6_wrangle_output,
    run_wrangling_comparison,
    out_format,
):
    input_dir = test_cmip6_crunch_output
    output_dir = str(tmpdir)

    runner = CliRunner()
    result = runner.invoke(
        wrangle_netcdf_scm_ncs,
        [
            input_dir,
            output_dir,
            "cmip6output wrangling regression test",
            "--drs",
            "CMIP6Output",
            "-f",
            "--number-workers",
            1,
            "--out-format",
            out_format,
            # have to avoid files which only have a single year and will fail with
            # average wrangling
            "--regexp",
            "^(?!.*(CESM2/historical/r7i1p1f1|IPSL-CM6A-LR/historical/r1i1p1f1)).*(/tas/|/tos/|/rlut/|/lai/).*$",
        ],
    )
    assert result.exit_code == 0, result.output

    run_wrangling_comparison(
        join(output_dir, "CMIP6"),
        join(test_cmip6_wrangle_output, out_format, "CMIP6"),
        update=update_expected_files,
    )



"""
Miscellaneous definitions used in netCDF-SCM
"""

_SCM_TIMESERIES_META_COLUMNS = [
    "variable",
    "variable_standard_name",
    "region",
    "scenario",
    "climate_model",
    "member_id",
    "mip_era",
    "activity_id",
]
"""Metadata columns to include when creating SCM timeseries"""

_LAND_FRACTION_REGIONS = [
    "World",
    "World|Land",
    "World|Ocean",
    "World|Northern Hemisphere",
    "World|Northern Hemisphere|Land",
    "World|Northern Hemisphere|Ocean",
    "World|Southern Hemisphere",
    "World|Southern Hemisphere|Land",
    "World|Southern Hemisphere|Ocean",
]
"""
list: Regions required to perform land fraction calculations

We require all the ocean regions too as it only makes sense to return land fraction if
we're actually looking at data which contains a split.
"""

NAME_COMPONENTS_SEPARATOR = "_"
"""
str: Character assumed to separate different components within a filename

For example, if we come across a filename like 'tas_r1i1p1f1_UoM-Fancy' then
we assume that 'tas', 'r1i1p1f1' and 'UoM-Fancy' all refer to different bits
of metadata which are encoded within the filename.
"""

OUTPUT_PREFIX = "netcdf-scm"
"""
str: Prefix attached to outputs from netCDF-SCM by default
"""



"""
Helper tools for citing CMIP5 of the `Coupled Model Intercomparison Project <https://www.wcrp-climate.org/wgcm-cmip>`_ data
"""
import functools
import os.path

import pandas as pd
import requests
import tqdm

from ..errors import NonStandardLicenseError


def check_license_cmip5(scmdata):
    """
    Check the license information of a CMIP5 dataset.

    This uses the information found at
    `<https://pcmdi.llnl.gov/mips/cmip5/docs/CMIP5_modeling_groups.pdf>`_.

    Parameters
    ----------
    scmdata : :obj:`scmdata.ScmRun`
        Dataset of which to check the license information

    Raises
    ------
    NonStandardLicenseError
        The dataset's license is non-standard i.e. is not a
        `Creative Commons Attribution ShareAlike 4.0 International License <https://creativecommons.org/licenses/by-sa/4.0/>`_
    """
    climate_model = scmdata.get_unique_meta("climate_model", no_duplicates=True)

    non_standard_model = any(
        [climate_model.startswith(v) for v in ("MIROC", "MRI", "NICAM")]
    )
    if non_standard_model:
        error_msg = (
            "Output is only available for non-commercial research and "
            "education purposes, see "
            "`<https://pcmdi.llnl.gov/mips/cmip5/docs/CMIP5_modeling_groups.pdf>`_"
        )
        raise NonStandardLicenseError(error_msg)


@functools.lru_cache()
def _query_cmip5_doi(doi):  # pragma: no cover
    headers = {
        "Accept": "application/x-bibtex",
    }

    response = requests.get(doi, headers=headers)
    response.raise_for_status()
    return response.text


def get_cmip5_bibtex_citation(model, scenario):
    """
    Get CMIP5 bibtex citation

    Parameters
    ----------
    model : str
        Model of interest

    scenario : str
        Scenario of interest

    Returns
    -------
    str
        Bibtex citation

    Raises
    ------
    ValueError
        No CMIP5 data is available for ``model`` ``scenario`` hence no
        citation can be provided.
    """
    cmip5_dois = _load_cmip5_dois().set_index(["climate_model", "scenario"])

    try:
        doi = cmip5_dois.loc[(model, scenario), "doi"].values[0]
    except KeyError:
        raise ValueError("No CMIP5 data for {} {}".format(model, scenario))

    citation_key = _get_cmip5_citation_key(model, scenario)

    bib_tex_citation = _query_cmip5_doi(doi).replace(doi.lower(), citation_key)

    return bib_tex_citation


def _get_cmip5_citation_key(model, scenario):
    return "cmip5data_{}_{}".format(model, scenario)


def _load_cmip5_dois():
    return pd.read_csv(
        os.path.join(
            os.path.dirname(__file__), "cmip5_climate-model_scenario_citations.csv"
        )
    )


def _determine_cmip5_modelling_group(model):  # pragma: no cover
    if model.startswith("ACCESS"):
        return "CSIRO-BOM"

    if model.startswith("BCC-"):
        return "BCC"

    if model.startswith("BESM"):
        return "INPE"

    if model.startswith("BNU-ESM"):
        return "GCESS"

    if model.startswith("Can"):
        return "CCCMA"

    if model.startswith("CCSM4(RSMAS)"):
        return "RSMAS"

    if model.startswith("CCSM4"):
        return "NCAR"

    if model.startswith("CESM1"):
        return "NSF-DOE-NCAR"

    if model.startswith("CFS"):
        return "COLA and NCEP"

    if model.startswith("CMCC"):
        return "CMCC"

    if model.startswith("CNRM"):
        return "CNRM-CERFACS"

    if model.startswith("CSIRO-Mk"):
        return "CSIRO-QCCCE"

    if model.startswith("EC-EARTH"):
        return "EC-EARTH"

    if model.startswith("FGOALS-g2"):
        return "LASG-CESS"

    if model.startswith("FGOALS-"):
        return "LASG-IAP"

    if model.startswith("FIO-"):
        return "FIO"

    if model.startswith("GEOS-5"):
        return "NASA GMAO"

    if model.startswith("GFDL-"):
        return "NOAA GFDL"

    if model.startswith("GISS-"):
        return "NASA GISS"

    if model.startswith("HadGEM2-AO"):
        return "NIMR/KMA"

    if model.startswith("HadCM3") or model.startswith("HadGEM2"):
        return "MOHC"

    if model.startswith("INM-CM4") or model.startswith("INMCM4"):
        return "INM"

    if model.startswith("IPSL-CM5"):
        return "IPSL"

    if model.startswith("MIROC"):
        return "MIROC"

    if model.startswith("MPI-ESM"):
        return "MPI-M"

    if model.startswith("MRI-"):
        return "MRI"

    if model.startswith("NICAM"):
        return "NICAM"

    if model.startswith("NorESM1"):
        return "NCC"

    raise NotImplementedError(model)


def _clean_ar5_html():  # pragma: no cover
    from bs4 import BeautifulSoup

    with open(
        os.path.join(os.path.dirname(__file__), "data", "ar5-ref-info.html")
    ) as fh:
        soup = BeautifulSoup(fh.read(), "lxml")

    model_doi_table = soup.find("table", attrs={"id": "customers"})
    model_doi_table_data = model_doi_table.tbody.find_all("tr")

    data = {
        "model": [],
        "scenario": [],
        "doi": [],
    }
    for i, tr in tqdm.tqdm(enumerate(model_doi_table_data)):
        if i == 0:
            headers = [th.text.replace("\n", " ").strip() for th in tr.find_all("th")]
            model_col = headers.index("Model")

        elif i > 1:
            cols = tr.find_all("td")
            if not cols:
                model_doi_table_data[i - 1]
                continue

            if len(cols) == len(headers):
                model = cols[model_col].text.replace("\n", " ").strip()
            else:
                model = cols[0].text.replace("\n", " ").strip()

            model = model.replace("(Journal)", "").replace("\n", " ").strip()
            print(model)

            for j, col in enumerate(cols):
                if j <= model_col:
                    continue

                links = col.find_all("a")
                for link in links:
                    experiment = link.text.replace("\n", " ").strip()

                    try:
                        colour = link.span.get("style")
                        if not colour.startswith("color:darkgreen"):
                            if colour.startswith("color:LightSlateGrey"):
                                continue

                            if colour.startswith("color:darkred"):
                                continue

                            raise ValueError(colour)

                    except AttributeError:
                        continue

                    doi = link.get("href").replace("\n", " ").strip()

                    data["model"].append(model)
                    data["scenario"].append(experiment)
                    data["doi"].append(doi)

    data = pd.DataFrame(data)
    data.to_csv(
        os.path.join(
            os.path.dirname(__file__), "cmip5_climate-model_scenario_citations.csv"
        ),
        index=False,
    )


if __name__ == "__main__":
    _clean_ar5_html()



import re
from unittest.mock import patch

import pandas as pd
import pandas.testing as pdt
import pytest
from scmdata.run import ScmRun

from netcdf_scm.citing import check_licenses, get_citation_tables
from netcdf_scm.errors import NonStandardLicenseError


def make_scmrun(cm, mip_era):
    cols = {
        "climate_model": cm,
        "mip_era": mip_era,
        "model": "unspecified",
        "scenario": "unspecified",
        "variable": "tas",
        "unit": "K",
        "region": "World",
    }

    if mip_era == "CMIP6":
        cols["activity_id"] = "ScenarioMIP"
        cols["member_id"] = "r1i1p1f1"

    out = ScmRun(data=[1, 2, 3], index=[3, 4, 5], columns=cols)
    out.metadata = {"institution_id": "{}-institute".format(cm)}

    return out


@patch("netcdf_scm.citing.check_license_cmip6")
def test_check_licenses(mock_check_license_cmip6, caplog):
    def cmip6_check(inp):
        if inp.get_unique_meta("climate_model", no_duplicates=True).startswith("Nor"):
            raise NonStandardLicenseError

    mock_check_license_cmip6.side_effect = cmip6_check

    inp = [
        make_scmrun("ACCESS1.0", "CMIP5"),
        make_scmrun("MIROC-ESM", "CMIP5"),
        make_scmrun("ACCESS-CM2", "CMIP6"),
        make_scmrun("NorESM1", "CMIP6"),
    ]
    with pytest.warns(UserWarning) as record:
        res = check_licenses(inp)

    assert len(record) == 2

    assert str(record.list[0].message).startswith(
        "Non-standard license for MIROC-ESM\nLicense terms:"
    )
    assert str(record.list[1].message).startswith(
        "Non-standard license for NorESM1\nLicense terms:"
    )

    for r in res:
        assert r.get_unique_meta("climate_model", no_duplicates=True) in [
            "NorESM1",
            "MIROC-ESM",
        ]


@patch("netcdf_scm.citing.get_cmip5_bibtex_citation")
@patch("netcdf_scm.citing.get_cmip6_bibtex_citation")
def test_get_citation_tables(
    mock_get_cmip6_bibtex_citation, mock_get_cmip5_bibtex_citation
):
    def cmip5_bibtex_citation(model, scenario):
        if model.startswith("MIROC"):
            return "MIROC-ESM CMIP5 {} bibtex".format(scenario)

        return "ACCESS CMIP5 {} bibtex".format(scenario)

    mock_get_cmip5_bibtex_citation.side_effect = cmip5_bibtex_citation

    def cmip6_bibtex_citation(
        mip_era, activity_id, institute_id, model, scenario, version
    ):
        if model.startswith("Nor"):
            return "Nor CMIP6 {} {} bibtex".format(scenario, version)

        return "ACCESS CMIP6 {} {} bibtex".format(scenario, version)

    mock_get_cmip6_bibtex_citation.side_effect = cmip6_bibtex_citation

    access_cmip5 = make_scmrun("ACCESS1.0", "CMIP5")
    access_cmip5.metadata["(child) crunch_source_files"] = (
        "Files: ["
        "'/cmip5/ssp585/Amon/tas/ACCESS1.0/r1i1p1/tas_Amon_ACCESS1.0_ssp585_r1i1p1_200501-210012.nc'"
        "]"
    )
    access_cmip5.metadata["(parent) crunch_source_files"] = (
        "Files: ["
        "'/cmip5/historical/Amon/tas/ACCESS1.0/r1i1p1/tas_Amon_ACCESS1.0_historical_r1i1p1_185001-200412.nc'"
        "]"
    )

    miroc_cmip5 = make_scmrun("MIROC-ESM", "CMIP5")
    miroc_cmip5.metadata["crunch_source_files"] = (
        "Files: ["
        "'/cmip5/ssp585/Amon/tas/MIROC-ESM/r1i1p2/tas_Amon_MIROC-ESM_ssp585_r1i1p2_200501-230012.nc'"
        "]"
    )

    access_cmip6 = make_scmrun("ACCESS-CM2", "CMIP6")
    access_cmip6.metadata["crunch_source_files"] = (
        "Files: ["
        "'/CMIP6/CMIP/ACCESS-CM2-institute/ACCESS-CM2/historical/r1i1p1f1/Amon/tas/gn/v20190627/tas_Amon_ACCESS-CM2_historical_r1i1p1f1_gn_185001-194912.nc', "
        "'/CMIP6/CMIP/ACCESS-CM2-institute/ACCESS-CM2/historical/r1i1p1f1/Amon/tas/gn/v20190627/tas_Amon_ACCESS-CM2_historical_r1i1p1f1_gn_195001-201412.nc'"
        "]"
    )

    noresm_cmip6 = make_scmrun("NorESM1", "CMIP6")
    noresm_cmip6.metadata["(child) crunch_source_files"] = (
        "Files: ["
        "'/CMIP6/CMIP/NorESM1-institute/NorESM1/G6solar/r2i1p1f1/Amon/tas/gn/v20200101/tas_Amon_NorESM1_G6solar_r2i1p1f1_gn_201501-205012.nc'"
        "]"
    )
    noresm_cmip6.metadata["(parent) crunch_source_files"] = (
        "Files: ["
        "'/CMIP6/CMIP/NorESM1-institute/NorESM1/ssp585/r2i1p1f1/Amon/tas/gn/v20200101/tas_Amon_NorESM1_ssp585_r2i1p1f1_gn_201501-205012.nc'"
        "]"
    )
    noresm_cmip6.metadata["(grandparent) crunch_source_files"] = (
        "Files: ["
        "'/CMIP6/CMIP/NorESM1-institute/NorESM1/historical/r2i1p1f1/Amon/tas/gn/v20200101/tas_Amon_NorESM1_historical_r2i1p1f1_gn_185001-200412.nc'"
        "]"
    )
    noresm_cmip6.metadata["(greatgrandparent) crunch_source_files"] = (
        "Files: ["
        "'/CMIP6/CMIP/NorESM1-institute/NorESM1/piControl/r2i1p1f1/Amon/tas/gn/v20200101/tas_Amon_NorESM1_piControl_r2i1p1f1_gn_000101-010012.nc'"
        "]"
    )

    inp = [
        access_cmip5,
        miroc_cmip5,
        access_cmip6,
        noresm_cmip6,
    ]

    cmip5_references_exp = (
        pd.DataFrame(
            {
                "Modelling Group": ["CSIRO-BOM", "CSIRO-BOM", "MIROC"],
                "Climate Model": ["ACCESS1.0", "ACCESS1.0", "MIROC-ESM"],
                "Scenario": ["historical", "ssp585", "ssp585"],
                "Reference": [
                    r"\citet{cmip5data_ACCESS1.0_historical}",
                    r"\citet{cmip5data_ACCESS1.0_ssp585}",
                    r"\citet{cmip5data_MIROC-ESM_ssp585}",
                ],
            }
        )
        .set_index(["Modelling Group", "Climate Model", "Scenario"])
        .sort_index()
    )

    cmip6_references_exp = (
        pd.DataFrame(
            {
                "Modelling Group": [
                    "ACCESS-CM2-institute",
                    "NorESM1-institute",
                    "NorESM1-institute",
                    "NorESM1-institute",
                    "NorESM1-institute",
                ],
                "Climate Model": [
                    "ACCESS-CM2",
                    "NorESM1",
                    "NorESM1",
                    "NorESM1",
                    "NorESM1",
                ],
                "Scenario": [
                    "historical",
                    "G6solar",
                    "historical",
                    "piControl",
                    "ssp585",
                ],
                "Reference": [
                    r"\citet{cmip6data_CMIP6_CMIP_ACCESS-CM2-institute_ACCESS-CM2_historical}",
                    r"\citet{cmip6data_CMIP6_CMIP_NorESM1-institute_NorESM1_G6solar}",
                    r"\citet{cmip6data_CMIP6_CMIP_NorESM1-institute_NorESM1_historical}",
                    r"\citet{cmip6data_CMIP6_CMIP_NorESM1-institute_NorESM1_piControl}",
                    r"\citet{cmip6data_CMIP6_CMIP_NorESM1-institute_NorESM1_ssp585}",
                ],
            }
        )
        .set_index(["Modelling Group", "Climate Model", "Scenario"])
        .sort_index()
    )

    res = get_citation_tables(inp)

    pdt.assert_frame_equal(res["cmip5_references"], cmip5_references_exp)
    assert set(res["cmip5_bibtex"]) == set(
        [
            "ACCESS CMIP5 historical bibtex",
            "ACCESS CMIP5 ssp585 bibtex",
            "MIROC-ESM CMIP5 ssp585 bibtex",
        ]
    )

    pdt.assert_frame_equal(res["cmip6_references"], cmip6_references_exp)
    assert set(res["cmip6_bibtex"]) == set(
        [
            "Nor CMIP6 piControl v20200101 bibtex",
            "Nor CMIP6 historical v20200101 bibtex",
            "Nor CMIP6 ssp585 v20200101 bibtex",
            "Nor CMIP6 G6solar v20200101 bibtex",
            "ACCESS CMIP6 historical v20190627 bibtex",
        ]
    )


def test_get_citation_tables_no_dependencies():
    miroc_run = make_scmrun("MIROC-ESM", "CMIP5")
    inp = [
        miroc_run,
        make_scmrun("NorESM1", "CMIP6"),
    ]

    error_msg = re.escape("Could not find any dependencies for {}".format(miroc_run))
    with pytest.raises(ValueError, match=error_msg):
        get_citation_tables(inp)


def test_get_citation_tables_bad_mip_era():
    inp = [
        make_scmrun("ACCESS1.0", "dud mip"),
    ]

    error_msg = "Cannot generate citation table for mip_era: dud mip"
    with pytest.raises(ValueError, match=error_msg):
        get_citation_tables(inp)
